/* NFIDENT: a program for neuro-fuzzy function approximation

   Copyright (C) 1997-2000 

   Detlef Nauck (nauck@ieee.org, 
   http:\\fuzzy.cs.uni-magdeburg.de/~nauck)  
   hereby disclaims all copyright interests in the program 
   NFIDENT written by Detlef Nauck

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 1, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/* anfis.c */
/* This file contains functions to train a TSK fuzzy system by the */
/* ANFIS learning algorithm */

#include <math.h>
#include <stdio.h>
#include "global.h"
#include "tsk.h"
#include "lmse.h"

/* Reset_Delta is a local function that copies all delta values of a */
/* TSK fuzzy system to its old delta values, and then resets them. */
/* Parameters: TSK: pointer to a TSK fuzzy system */
void Anfis_Reset_Delta (TSKFuzzySystem *TSK, LParam *LP)
{
  int i, j, k;
  FSet *F;

  /* reset delta values of input fuzzy sets */
  for (i = 0; i < TSK->in; i++)
    {
      F = TSK->invars[i]->fsets;
      for (j = 0; j < TSK->invars[i]->count; j++)
	{
	  F[j].od_a = F[j].d_a; F[j].od_b = F[j].d_b; 
	  F[j].od_c = F[j].d_c; F[j].od_d = F[j].d_d;
	  F[j].d_a = 0; F[j].d_b = 0; F[j].d_c = 0; F[j].d_d = 0;
	};
    };
  /* reset delta values of conclusion parameters */
  if (!LP->uselmse)
    {
      for (i = 0; i < TSK->rules; i++)
	{
	  for (j = 0; j < TSK->out; j++)
	    {
	      for (k = 0; k < TSK->in + 1; k++)
		{
		  TSK->rulebase[i]->old_delta[j][k] = 
		    TSK->rulebase[i]->delta[j][k];
		  TSK->rulebase[i]->delta[j][k] = 0;
		};
	    };
	};
    };

}; /* Reset Delta */


/* Compute the delta values for conclusion and antecedent parameters */
void Anfis_Compute_Delta (TSKFuzzySystem *TSK, double *P, LParam *LP)
{
  int i, j, k;
  double d;
  double d_a, d_b, d_c, d_d;
  PTSKRule rule;
  FSet *F;
  
  for (i = 0; i < TSK->rules; i++)
    {
      /* printf ("Conclusion, rule %i.\n", i); */
      rule = TSK->rulebase[i];
      if (rule->hit > 0.001)
	{
	  rule->error = 0;
	  /* first the deltas of the conclusion parameters */
	  for (j = 0; j < TSK->out; j++)
	    { 
	      d = TSK->errors[j] * rule->normhit;

	      if (!LP->uselmse)
		{ /* compute output deltas, if we don't use LMSE */
		  for (k = 0; k < TSK->in + 1; k++)
		    {
		      rule->delta[j][k] += d * ((k < TSK->in)?P[k]:1.0);
		    }; /* for k*/
		};

	      /* rule delta for the antecedent deltas */
	      rule->error += TSK->errors[j] * 
		(rule->output[j] - TSK->outputs[j]) 
		* (rule->normhit / rule->hit);
	    };/* for j*/

	  /* printf ("rule %i, error: %f.\n", i, rule->error); */
	  /* now the deltas of the antecedent parameters */
	  for (j = 0; j < TSK->in; j++)
	    {
	      F = &(TSK->invars[j]->fsets[rule->antecedent[j]]);
	      if (F->hit > 0.001) /* "if" is not necessary, if a t-norm is used */
		{             /* to compute the rule's fulfillment */
		  d = rule->error / F->hit;

		  switch (F->form)
		    {
		    case TRIANGLE:
		      if ((F->a < P[j]) && (P[j] < F->b))
			{
			  d_a = LP->sigma_a * d 
			    * (-1.0 - F->hit / (F->b - F->a));
			  d_b = LP->sigma_b * d * 
			    (- (P[j] - F->a) / sqr((F->b - F->a)));
			  if (!LP->asym)
			    {
			      F->d_c += -d_a + d_b;
			      F->d_a += d_b;
			    };
			  F->d_a += d_a;
			  F->d_b += d_b;
			}/* if */
		      else
			{
			  if ((F->b < P[j]) && (P[j] < F->c))
			    {
			      d_b = LP->sigma_b * d * 
				((F->c - P[j]) / sqr((F->c - F->b)));
			      d_c = d * LP->sigma_c * 
				((1.0 - F->hit) / (F->c - F->b));
			      if (!LP->asym)
				{
				  F->d_a += -d_c + d_b;
				  F->d_c += d_b;
				};
			      F->d_c += d_c;
			      F->d_b += d_b;
			    };
			}; /* else */
		      break;
		    case TRAPEZOID:
		      if ((F->a < P[j]) && (P[j] < F->b))
			{
			  d_a = d * LP->sigma_a * 
			    ((-1.0 - F->hit) / (F->b - F->a));
			  d_b = d * LP->sigma_b * 
			    (- F->hit / (F->b - F->a));
			  if (!LP->asym)
			    {
			      F->d_d += -d_a + d_b;
			      F->d_a += d_b;
			    };
			  F->d_a += d_a;
			  F->d_b += d_b;
			}/* if */
		      else
			{
			  if ((F->c < P[j]) && (P[j] < F->d))
			    {
			      d_c = d * LP->sigma_c * 
				(F->hit / (F->d - F->c));
			      d_c = d * LP->sigma_c * 
				((1.0 - F->hit) / (F->d - F->c));
			      if (!LP->asym)
				{
				  F->d_a += -d_d + d_b;
				  F->d_d += d_b;
				};
			      F->d_d += d_d;
			      F->d_c += d_c;
			    }
			}; /* else */
		      break;
		    case GAUSSIAN:
		      /*d_b = ((-2.0 * sqr(GF) * (F->b - P[j])) / sqr(F->a))
			    * F->hit;
		      F->d_a += ((-d_b * (F->b - P[j])) / F->a) * d * 
			        LP->sigma_a;*/
		      /*d_b = ((-2.0 * sqr(GF) * (F->b - P[j]))/ sqr(F->a))
			    * rule->error;
		      F->d_a += ((-d_b * (F->b - P[j]))/ F->a) * LP->sigma_a + 
			        LP->momentum * F->od_a;
		      F->d_b += d_b * LP->sigma_b +
			        LP->momentum * F->od_b;*/
		      d_b = (P[j] - F->b) * 
			sgn(P[TSK->in+j] - rule->output[j]) * 
			rule->hit * F->hit;
		      F->d_a += -d_b * LP->sigma_a +  LP->momentum * F->od_a;
		      F->d_b += d_b * LP->sigma_b +  LP->momentum * F->od_b;
		      break;
		    }; /* switch */
		}; /* if (F->hit > 0) */
	    }; /* for j */
	}; /* if (rule->hit > 0) */
    }; /* for i*/

}; /* Compute Delta */


void Anfis_Adapt (TSKFuzzySystem *TSK, LParam *LP, int norm)
{
  int i, j, k;
  PTSKRule rule;

  /* adapt membership functions (learning rate already applied) */
  for (i = 0; i < TSK->in; i++)
    {
      Adapt_LingVar (TSK->invars[i], norm, LP->nopass, LP->at05);
    }; /* for i*/

  if (!LP->uselmse)
    { /* adapt conclusions, learning rate and momentum must still be applied */
      for (i = 0; i < TSK->rules; i++)
	{
	  rule = TSK->rulebase[i];
	  for (j = 0; j < TSK->out; j++)
	    {
	      for (k = 0; k < TSK->in + 1; k++)
		{
		  /* Note that we change rule->delta also! */
		  rule->conclusion[j][k] += 
		    (rule->delta[j][k] *= (LP->sigma_t / norm))
		    +(LP->momentum * rule->old_delta[j][k]);
		}; /* for k */
	    }; /* for j */
	}; /* for i*/
    };
}; /* Adapt */

int Anfis_Training (TSKFuzzySystem *TSK, PSet *P, LParam *LP, PSet *G)
{
  int i;
  int history;
  int stop;
  int epoch;
  int last_improve;
  double paterror;
  double train_error;
  double last_error;
  double best_error;
  TSKFuzzySystem *Save;
  FILE *logfile;
  char savefile[80];

  stop = 0;
  epoch = 0; 
  last_improve = 0;
  history = 0;
  train_error = 0;
  last_error = 0;
  Save = NULL;

  if ((logfile = fopen (LP->logfilename, "w")) == NULL)
    {
      printf ("Cannot create logfile %s!\n", LP->logfilename);
      return -1;
    };
  
  Anfis_Reset_Delta(TSK, LP);
  Anfis_Reset_Delta(TSK, LP); /* two times, to also reset old delta values */

  while (!stop)
    {
      epoch++;
      TSK->Error = 0;
      Copy_TSK (TSK, &Save, LP);



      if (LP->uselmse)
	{
	  for (i = 0; i < P->count; i++)
	    {
	      TSK_Propagate_Pattern (TSK, P, i);
	      /*LMSE_Step (P->Pats[i]);*/
	    };
	};

      for (i = 0; i < P->count; i++)
	{
	  TSK_Propagate_Pattern (TSK, P, i);

	  Anfis_Compute_Delta (TSK, P->Pats[i], LP);

	  /* do a learning step, if not in batch mode */
	  if (!LP->batch)
	    {
	      Anfis_Adapt (TSK, LP, 1);
	      Anfis_Reset_Delta (TSK, LP);
	    };
	};
 
      TSK->Error = sqrt (TSK->Error / (P->count * TSK->out));

      /* if in batch mode, do a learning step now */
      if (LP->batch)
	{
	  Anfis_Adapt (TSK, LP, P->count);
	  Anfis_Reset_Delta(TSK, LP);
	};

      train_error = TSK->Error;

      if (G != NULL)
	{ /* if generalization patterns are given, do cross-validation */
	  TSK->Error = 0;
	  for (i = 0; i < G->count; i++)
	    {
	      TSK_Propagate_Pattern (TSK, G, i);
	    };
	  TSK->Error = sqrt (TSK->Error / (P->count * TSK->out));
 	};

      /* store best error value, and the epoch it occured in, and save net */
      if (epoch == 1)
	{
	  best_error = TSK->Error;
	  last_error = TSK->Error;
	  last_improve = 1;
	  Copy_TSK (Save, &BestAnfis, LP);
	}
      else
	{
	  if (best_error > TSK->Error)
	    {
	      best_error = TSK->Error;
	      last_improve = epoch;
	      if (G != NULL)
		Copy_TSK (TSK, &BestAnfis, LP);
	      else
		Copy_TSK (Save, &BestAnfis, LP);
	    };
	  if (last_error <  TSK->Error)
	    {
	      history <<=1;
	      history++;
	    }
	  else
	    {
	      history <<=1;
	    };
	  
	  history &= 0xF;
	  
	  if (history == 0)
	    {
	      LP->sigma_a *= 1.1;
	      LP->sigma_b *= 1.1;
	      LP->sigma_c *= 1.1;
	      LP->sigma_d *= 1.1;
	      LP->sigma_t *= 1.1;
	      printf ("Learning rate increased in epoch %i, sigma_a = %f (%i).\n", 
		  epoch, LP->sigma_a, history);
	    };
	  if ((history == 3) || (history == 5) || (history == 7) ||
	      (history == 11)|| (history == 15)) 
	    {
	      LP->sigma_a *= 0.9;
	      LP->sigma_b *= 0.9;
	      LP->sigma_c *= 0.9;
	      LP->sigma_d *= 0.9;
	      LP->sigma_t *= 0.9;
	      printf ("Learning rate decreased in epoch %i, sigma_a = %f (%i).\n", 
		      epoch, LP->sigma_a, history);
	    };
	};
      last_error = TSK->Error;

      /* check, whether we must stop now */
      stop = (((TSK->Error <= LP->minerror) || (epoch > LP->maxepoch) ||
	      (epoch - last_improve > LP->chgepoch)) && 
	     (epoch > LP->minepoch)) || StopOnSignal;
      if (((epoch % LP->logepoch) == 0) || (epoch == 1) || stop)
	{
	  if (G != NULL)
	    {
	      printf ("Epoch %i: training error: %f, ", epoch, train_error);
	      printf ("generalization error: %f,\n", TSK->Error);
	      printf ("best error of %f in epoch %i.\n", 
		      best_error, last_improve);
	    }
	  else
	    {
	      printf ("Epoch %i: Error = %f, best error of %f in epoch %i.\n",
		      epoch, TSK->Error, best_error, last_improve);
	    };
	  fprintf (logfile, "%i %f %f %i\n", 
		   epoch, TSK->Error, best_error, last_improve);
	  strcpy (savefile, LP->savenetfilename);
	  Build_Filename (savefile, epoch);
	  Write_TSKFuzzy_System (savefile, BestAnfis);
	  Write_TSKFuzzy_System ("current.nfs", Anfis);
	};
    };
  fclose (logfile);
  return epoch;
};/* Anfis Training */

