/* NFIDENT: a program for neuro-fuzzy function approximation

   Copyright (C) 1997-2000 

   Detlef Nauck (nauck@ieee.org, 
   http:\\fuzzy.cs.uni-magdeburg.de/~nauck)  
   hereby disclaims all copyright interests in the program 
   NFIDENT written by Detlef Nauck

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 1, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <math.h>
#include <stdio.h>
#include "global.h"
#include "fuzzy.h"

/* COG: defuzzification by center of gravity */
/* corrected on 07/10/00, errors reported by Harald Zainzinger */
/* added missing breaks and modified aggregation in inner loop */
/* now each x is counted only once, i.e. we have true COA */
/* instead of simplified COG */
double COG (LingVar *lvar, TNORM tnorm, SNORM snorm)
{
  double nominator;
  double denominator;
  double intervall;
  double x;
  double my;
  double t;
  int   i;

  intervall = (lvar->upper - lvar->lower) / SAMPLES;
  nominator = 0; 
  denominator = 0;

  /* scan the whole output range */
  for (x = lvar->lower; x <= lvar->upper; x += intervall)
    {
      t = 0;
      for (i = 0; i < lvar->count; i++) /* for all fuzzy sets */
	{
	  if (lvar->fsets[i].hit > 0)
	    {
	      switch (lvar->fsets[i].form)
		{
		case TRIANGLE: 
		  {
		    if ((x >= lvar->fsets[i].a) && (x <= lvar->fsets[i].c))
		      {
			my = tnorm(lvar->fsets[i].hit, 
				   MS_Degree(x, &(lvar->fsets[i])));
			t = snorm(t,my);
		      };
		  };
		  break;
		case TRAPEZOID:
		  {
		    if ((x >= lvar->fsets[i].a) && (x <= lvar->fsets[i].d))
		      {
			my = tnorm(lvar->fsets[i].hit, 
				   MS_Degree(x, &(lvar->fsets[i])));
			t = snorm(t,my);
		      };		
		  };
		  break;
		case GAUSSIAN:
		  {
		    my = tnorm(lvar->fsets[i].hit, 
			      MS_Degree(x, &(lvar->fsets[i])));
		    t = snorm(t,my);
		  };
		  break;
		}; /* switch */
	    };/* if */
	}; /* for */
	nominator += x * t;
	denominator += t;
    }; /* for */

  if (denominator > 0)
    return (nominator / denominator);
  else
    return 0;

} /* COG */

/* MOM: defuzzification by mean of maxima */

/* 07/10/00: added SNORM because of updated COG function */
/* both function must have the same parameters */
/* however, here we don't need the SNORM and simply ignore it */
double MOM (LingVar *lvar, TNORM tnorm, SNORM snorm)
{

  double nominator;
  double denominator;
  int i;

  nominator = 0;
  denominator = 0;

  for (i = 0; i < lvar->count; i++)
    {
      if (lvar->fsets[i].hit > 0)
	{
	  switch (lvar->fsets[i].form)
	    {
	    case TRIANGLE: 
	      if (lvar->fsets[i].ls)
		nominator += lvar->fsets[i].hit * ((lvar->fsets[i].a +
			      lvar->fsets[i].hit * (lvar->fsets[i].b
			      - lvar->fsets[i].c) + lvar->fsets[i].c) / 2);
	      else
		if (lvar->fsets[i].rs)
		  nominator += lvar->fsets[i].hit * ((lvar->fsets[i].c +
			      lvar->fsets[i].hit * (lvar->fsets[i].b
			      - lvar->fsets[i].a) + lvar->fsets[i].a) / 2);
	        else
		  nominator += lvar->fsets[i].hit * ((lvar->fsets[i].hit * 
			       (2 * lvar->fsets[i].b - lvar->fsets[i].c - 
			        lvar->fsets[i].a) + 
			       lvar->fsets[i].a + lvar->fsets[i].c) / 2);
	      break;
	    case TRAPEZOID:
	      if (lvar->fsets[i].ls)
		nominator += lvar->fsets[i].hit * ((lvar->fsets[i].b +
			      lvar->fsets[i].hit * (lvar->fsets[i].c
			      - lvar->fsets[i].d) + lvar->fsets[i].d) / 2);
	      else
		if (lvar->fsets[i].rs)
		  nominator += lvar->fsets[i].hit * ((lvar->fsets[i].c +
			      lvar->fsets[i].hit * (lvar->fsets[i].b
			      - lvar->fsets[i].a) + lvar->fsets[i].a) / 2);
	        else
		  nominator += lvar->fsets[i].hit * ((lvar->fsets[i].hit * 
			       (lvar->fsets[i].b + lvar->fsets[i].c - 
			        lvar->fsets[i].a - lvar->fsets[i].d) + 
			       lvar->fsets[i].a + lvar->fsets[i].d) / 2);
	      break;
	    case GAUSSIAN:
	      if (lvar->fsets[i].ls)
		nominator += lvar->fsets[i].hit * 
		             (lvar->fsets[i].b + lvar->fsets[i].a * 
		              sqrt(-log(lvar->fsets[i].hit)) / (2*GF));
	      else
		if (lvar->fsets[i].rs)
		  nominator += lvar->fsets[i].hit * 
		               (lvar->fsets[i].b - lvar->fsets[i].a * 
		                sqrt(-log(lvar->fsets[i].hit)) / (2*GF));
	        else
		  nominator += lvar->fsets[i].hit * lvar->fsets[i].b;
	      break;
	    };
	  denominator += lvar->fsets[i].hit;
	};
    };

  if (denominator > 0)
    return (nominator / denominator);
  else
    return 0;

}; /* MOM */
