/* NFIDENT: a program for neuro-fuzzy function approximation

   Copyright (C) 1997-2000 

   Detlef Nauck (nauck@ieee.org, 
   http:\\fuzzy.cs.uni-magdeburg.de/~nauck)  
   hereby disclaims all copyright interests in the program 
   NFIDENT written by Detlef Nauck

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 1, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <math.h>
#include <stdio.h>
#include <float.h>
#include "global.h"

/* MS_Degree: calculate membership degree of x */
double MS_Degree (double x, FSet *mu)
{
  double hit;
  double tmp;

  switch (mu->form)
    {
    case TRIANGLE:
      {
	if (((x < mu->b) && mu->ls) || ((x > mu->b) && (mu->rs)))
	  hit = 1.0;
	else
	  {
	    if ((x >= mu->b) && (x < mu->c))
	      hit = ((mu->c-x)/(mu->c-mu->b));
	    else
	      {
		if ((x < mu->b) && (x > mu->a))
		  hit = ((x-mu->a)/(mu->b-mu->a));
		else
		  hit =  0.0;
	      };
	  };
	break;
      };
    case TRAPEZOID:
      {
	if (((x < mu->b) && mu->ls) || ((x > mu->c) && (mu->rs)) || 
	    ((x >= mu->b) && (x <= mu->c)))
	  hit = 1.0;
	else
	  {
	    if ((x >= mu->c) && (x < mu->d))
	      hit = ((mu->d-x)/(mu->d-mu->c));
	    else
	      {
		if ((x < mu->b) && (x > mu->a))
		  hit = ((x-mu->a)/(mu->b-mu->a));
		else
		  hit = 0.0;
	      };
	  };
	break;
      };
    case GAUSSIAN:
      {
	if (((x < mu->b) && mu->ls) || ((x > mu->b) && (mu->rs)))
	  {
	    hit = 1.0;
	  }
	else
	  {
	    tmp = GF * (mu->b - x) / mu->a;
	    hit = exp (- sqr(tmp));
	  };
	break;
      };
    };
  return hit;
}; /* MS_Degree */

/* Make_Partition: Create membership functions for a variable */
int Make_Partition (LingVar *lvar)
{
 double width;
 double position;
 int i;

 if (lvar == NULL)
   {
     printf("NULL pointer in Make_Partition!");
     return -1;
   };

 width = (lvar->upper - lvar->lower);

 switch (lvar->form)
   {
   case TRIANGLE: width = width / (lvar->count + 1); break;
   case TRAPEZOID: width = width / (2 * lvar->count - 1); break;
   case GAUSSIAN: width = width / (lvar->count - 1); break;
   };

 position = lvar->lower;

 for (i = 0; i < lvar->count; i++)
   {
     lvar->fsets[i].form = lvar->form;
     lvar->fsets[i].hit = 0;
     switch (lvar->form)
       {
       case TRIANGLE:
	 {
	   lvar->fsets[i].a = position;
	   lvar->fsets[i].b = position+ width;
	   lvar->fsets[i].c = position+ 2 * width;
	   lvar->fsets[i].d = 0;
	   lvar->fsets[i].ls = (i == 0);
	   lvar->fsets[i].rs = (i == lvar->count-1);
	   sprintf (lvar->fsets[i].name,"msf%i", i);
	   position = lvar->fsets[i].b;
	   break;
	 };
       case TRAPEZOID:
	 {
	   lvar->fsets[i].a = position;
	   if (i == 0)
	     {
	       lvar->fsets[i].b = position;
	       lvar->fsets[i].c = position + width;
	       lvar->fsets[i].d = position + 2 * width;
	     }
	   else
	     {
	       lvar->fsets[i].b = position + width;
	       lvar->fsets[i].c = position + 2 * width;
	       lvar->fsets[i].d = position + 3 * width;
	     };
	   lvar->fsets[i].ls = (i == 0);
	   lvar->fsets[i].rs = (i == lvar->count - 1);
	   if (i == lvar->count - 1)
	       lvar->fsets[i].d = lvar->fsets[i].c;
	   sprintf (lvar->fsets[i].name,"msf%i", i);
	   position = lvar->fsets[i].c;
	   break;
	 };
       case GAUSSIAN:
	 {
	   lvar->fsets[i].a = width;
	   lvar->fsets[i].b = position;
	   lvar->fsets[i].c = 0;
	   lvar->fsets[i].d = 0;
	   lvar->fsets[i].ls = (i == 0);
	   lvar->fsets[i].rs = (i == lvar->count - 1);
	   sprintf (lvar->fsets[i].name,"msf%i", i); 
	   position = position + width;
	   break;
	 };
       default: 
	 printf ("Error: invalid MSF type.\n");
	 return -1;
       };
   };
 return 0;
}; /* Make_Partition */


/* Write_Partition_For_Gnuplot: Create gnuplot data and command files */
int Write_Partition_For_Gnuplot (const char *FileName, LingVar *lvar,
				 const char *mode)
  {
    FILE *gnufile;
    int i;
    double a,b,c,d;

    if (lvar == NULL)
      {
	printf ("NULL pointer in Write_Partition_For_Gnuplot!\n");
	return -1;
      };

    printf ("Writing partition for variable %s.\n", lvar->name);

    if ((gnufile = fopen (FileName, mode)) == NULL)
      {
	printf ("Error opening gnuplot file for partition\n");
	return -1;
      }
    else
      {
	fprintf (gnufile, 
		 "set title \"Fuzzy Partition of Variable %s\"\n",
		       lvar->name);
	fprintf (gnufile, "set samples 1000\n");
	fprintf (gnufile, "set yrange [0:1]\n");
        fprintf (gnufile, "set xrange [%f:%f]\n", lvar->lower,
		 lvar->upper);

	for (i = 0; i < lvar->count; i++)
	  {
	    a = lvar->fsets[i].a;
	    b = lvar->fsets[i].b;
	    c = lvar->fsets[i].c;
	    d = lvar->fsets[i].d;

	    switch (lvar->fsets[i].form)
	      {
	      case TRIANGLE:
		  fprintf (gnufile, "f%i", i);
		  if (lvar->fsets[i].ls)
		    fprintf (gnufile,  GNU_LSTRIANGLE, b,b,c,c,c,b);
		  else
		    if (lvar->fsets[i].rs)
		      fprintf (gnufile,  GNU_RSTRIANGLE, b,b,a,a,b,a);
		    else
		      fprintf (gnufile,  GNU_TRIANGLE, a,b,a,b,a,c,b,c,c,b);
		  break;

	      case TRAPEZOID:
		  fprintf (gnufile, "f%i", i);
		  if (lvar->fsets[i].ls)
		    fprintf (gnufile,  GNU_LSTRAPEZOID, c,c,d,d,d,c);
		  else
		    if (lvar->fsets[i].rs)
		      fprintf (gnufile,  GNU_RSTRAPEZOID, a,b,a,b,a,b);
		    else
		      fprintf (gnufile,  GNU_TRAPEZOID, 
			       a,b,a,b,a,b,c,c,d,d,d,c);
		  break;

	      case GAUSSIAN:
		  fprintf (gnufile, "f%i", i);
		  fprintf (gnufile, GNU_EXPMSF, GF,b,a);

		  break;
	    };
	  };
	fprintf (gnufile, "plot ");
	for (i = 0; i < lvar->count; i++)
	  {
	    fprintf (gnufile, "f%i(x) t \"%s\"", i, lvar->fsets[i].name);
	    
	    if (i < lvar->count-1)
	      fprintf (gnufile, ", ");
	  };

	fprintf (gnufile, "\n");
	fprintf (gnufile, "pause -1 \"Hit return to continue\"\n");
      };
    fclose (gnufile);
    return 0;
  }; /* Write_Partition_For_Gnuplot */


int Create_Linguistic_Variables (PLingVar *ILV, PLingVar *OLV, 
				 PSet *P, PSet *G, PSet *T, 
				 LParam *LP, int systype)
{
  int i,j;
  double w;

  j = 0;
  
  for (i = 0; i < P->in; i++)
    {
      if ((ILV[i] = (LingVar *) calloc (1, sizeof(LingVar))) == NULL)
	{
	  printf ("Failed to create %ith linguistic variable.\n", i);
	  return -(i+1);
	};
      
      sprintf (ILV[i]->name, "Input%i", i+1);
      ILV[i]->lower = P->pmin[i];
      ILV[i]->upper = P->pmax[i];

      if (G != NULL)
	{
	  if (ILV[i]->lower > G->pmin[i])
	    {
	      ILV[i]->lower = G->pmin[i];
	    };
	  if (ILV[i]->upper < G->pmax[i])
	    {
	      ILV[i]->upper = G->pmax[i];
	    };
	};

      if (T != NULL)
	{
	  if (ILV[i]->lower > T->pmin[i])
	    {
	      ILV[i]->lower = T->pmin[i];
	    };
	  if (ILV[i]->upper < T->pmax[i])
	    {
	      ILV[i]->upper = T->pmax[i];
	    };
	};

      ILV[i]->form = LP->iform[i];
      ILV[i]->count = LP->imsf[i];
      if (Make_Partition (ILV[i]) < 0)
	{
	  printf ("Failed to create partition for input variable %i.\n", i);
	  return -(i+1);
	};
    };

  if (systype == MAMDANI)
    {
      for (j = 0; j < P->out; j++)
	{
	  if ((OLV[j] = (LingVar *) calloc (1, sizeof(LingVar))) == NULL)
	    {
	      printf ("Failed to create %ith linguistic output variable.\n", 
		      j);
	      return -(i+j+1);
	    };
	  
	  w = (P->pmax[i+j] - P->pmin[i+j]) * LP->outstretch;
	  sprintf (OLV[j]->name, "Output%i", j+1);
	  OLV[j]->lower = P->pmin[i+j] - w;
	  OLV[j]->upper = P->pmax[i+j] + w;
  
	  if (G != NULL)
	    {
	      if (OLV[j]->lower > G->pmin[i+j] - w)
		{
		  OLV[j]->lower = G->pmin[i+j] - w;
		};
	      if (OLV[j]->upper < G->pmax[i+j] + w)
		{
		  OLV[j]->upper = G->pmax[i+j] + w;
		};
	    };
	  
	  if (T != NULL)
	    {
	      if (OLV[j]->lower > T->pmin[i+j] - w)
		{
		  OLV[j]->lower = T->pmin[i+j] - w;
		};
	      if (OLV[j]->upper < T->pmax[i+j] + w)
		{
		  OLV[j]->upper = T->pmax[i+j] + w;
		};
	    };

  
	  OLV[j]->form = LP->oform[j];
	  OLV[j]->count = LP->omsf[j];
	  if (Make_Partition (OLV[j]) < 0)
	    {
	      printf ("Failed to create partition for output variable %i.\n", 
		      j);
	      return -(i+j+1);
	    };
	};
    };
  return (i+j);
}; /* Create Linguistic Variables */

void Compute_Pattern_Memberships (PLingVar *LV, double *P, int from, int count)
{
  int i;
  int j;
  double x;

  for (i = 0; i < count; i++)
    {
      for (j = 0; j < LV[i]->count; j++) 
	{
	  x = MS_Degree (P[i+from], &(LV[i]->fsets[j]));
	  LV[i]->fsets[j].hit = x;
	};
    };
}; /* Compute Pattern Memberships */

/* Adapt_LingVar carries out the apaptions of the membership functions,   */
/* for a given linguistic variable, and takes care of the constraints     */
/* nopass and at05. In addition the fuzzy sets must stay within their     */
/* domain, and the parameters must stay consistent. norm is used to       */
/* normalize the modifications (1 for online mode, number of patterns for */
/* batch mode) */

void Adapt_LingVar (LingVar *LV, int norm, int nopass, int at05)
{
  int j;
  double a,b,c,d;
  FSet *F;

  F = LV->fsets;
  for (j = 0; j < LV->count; j++)
    {
      /*printf ("V%i, F%i: %f %f %f %f.\n", i, j, F[j].a, F[j].b, F[j].c, 
	F[j].d);*/
      F[j].d_a /= norm;
      F[j].d_b /= norm;
      F[j].d_c /= norm;
      F[j].d_d /= norm;
      /*printf ("V%i, F%i: %f %f %f %f.\n", i, j, F[j].d_a, F[j].d_b, 
	F[j].d_c, F[j].d_d);*/
      a = F[j].a + F[j].d_a;
      b = F[j].b + F[j].d_b;
      c = F[j].c + F[j].d_c;
      d = F[j].d + F[j].d_d;
      /* check domains */
      switch (F[j].form)
	{
	case TRIANGLE:
	  /* does lower <= a < b < c <= upper hold ? */
	  if ((a < LV->lower) || (a >= b) || F[j].ls)
	    {
	      F[j].d_a = 0;
	      a = F[j].a;
	    };
	  if ((c > LV->upper) || (c <= b) || F[j].rs)
	    {
	      F[j].d_c = 0;
	      c = F[j].c;
	    };
	  /* may we still shift, after checking d_a and d_b? */
	  if ((a >= b) || (c <= b))
	    {
	      F[j].d_b = 0;
	      b = F[j].b;
	    };
	  /* check nopass constraint */
	  if (nopass && (j > 0))
	    {
	      if (a < F[j-1].a) 
		{
		  F[j].d_a = F[j-1].a - F[j].a;
		  a = F[j].a + F[j].d_a;
		};
	      if (b < F[j-1].b) 
		{
		  F[j].d_b = F[j-1].b - F[j].b;
		  b = F[j].b + F[j].d_b;
		};
	      if (c < F[j-1].c) 
		{
		  F[j].d_c = F[j-1].c - F[j].c;
		  c = F[j].c + F[j].d_c;
		};
	    };
	  if (nopass && (j < LV->count - 1))
	    {
	      if (a > F[j+1].a) 
		{
		  F[j].d_a = F[j+1].a - F[j].a;
		  a = F[j].a + F[j].d_a;
		};
	      if (b > F[j+1].b) 
		{
		  F[j].d_b = F[j+1].b - F[j].b;
		  b = F[j].b + F[j].d_b;
		};
	      if (c > F[j+1].c) 
		{
		  F[j].d_c = F[j+1].c - F[j].c;
		  c = F[j].c + F[j].d_c;
		};
	    };
	  /* check at05 constraint */
	  /* not yet implemented */
	  break;
	case TRAPEZOID:
	  break;
	case GAUSSIAN:

	  if (isnan(a))
	    {
	      F[j].d_a = 0;
	      F[j].od_a = 0;
	    };
	  if (isnan(b))
	    {
	      F[j].d_b = 0;
	      F[j].od_b = 0;
	    };

	  if (a <= 0)
	    {
	      F[j].d_a = 0;
	    };
	  if (a > (LV->upper - LV->lower))
	    {
	      F[j].d_a = 0;
	    };
	  if (F[j].ls || F[j].rs)
	    {
	      F[j].d_b = 0;
	      b = F[j].b;
	    };
	  if (nopass && (j > 0))
	    {
	      if (b < F[j-1].b)
		{
		  F[j].d_b = F[j-1].b - F[j].b;
		};
	    };
	  if (nopass && (j < LV->count - 1))
	    {
	      if (b > F[j+1].b)
		{
		  F[j].d_b = F[j+1].b - F[j].b;
		};
	    };
	  break;
	}; /* switch */
      /* modify membership function */
      /*printf ("V%i, F%i: %f %f %f %f.\n", i, j, F[j].d_a, F[j].d_b, 
	F[j].d_c, F[j].d_d);*/
      F[j].a += F[j].d_a;
      F[j].b += F[j].d_b;
      F[j].c += F[j].d_c;
      F[j].d += F[j].d_d;
      /*printf ("V%i, F%i: %f %f %f %f.\n", i, j, F[j].a, F[j].b, F[j].c, 
	F[j].d);*/
    }; /* for j */ 
}; /* Adapt LingVar */
