/* NFIDENT: a program for neuro-fuzzy function approximation

Copyright (C) 1997-2000 

Detlef Nauck (nauck@ieee.org, 
http:\\fuzzy.cs.uni-magdeburg.de/~nauck)  
hereby disclaims all copyright interests in the program 
NFIDENT written by Detlef Nauck

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 1, or (at your option)
any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <math.h>
#include <stdio.h>
#include "global.h"
#include "defuzz.h"

// uncomment one or more of the following lines to see debug output
// this is not complete,yet , only debug output required during development was added
//#define DEBUG_LEVEL1 // use to show selected summary results like rule statistics
//#define DEBUG_LEVEL2 // use to show stats as they are updated
//#define DEBUG_LEVEL3 // use show entering and exiting methods


MFuzzySystem *Create_Mamdani_Fuzzy_System (int nin, int nout, LParam *LP)
{
#ifdef DEBUG_LEVEL3
	printf("Entering Create_Mamdani_Fuzzy_System");
#endif

	MFuzzySystem *M;
	int i;

	/* allocate memory for fuzzy system */
	if ((M = (MFuzzySystem *) calloc (1, sizeof(MFuzzySystem))) == NULL)
	{
		printf ("Cannot allocate memory for fuzzy system!\n");
		return NULL;
	};

	/* allocate memory for input and output variables */
	if ((M->invars = (PLingVar *) calloc (nin, sizeof (PLingVar))) == NULL)
	{
		printf ("Cannot allocate memory for input variables!\n");
		return NULL;
	};

	if ((M->outvars = (PLingVar *) calloc (nout, sizeof (PLingVar))) == NULL)
	{
		printf ("Cannot allocate memory for output variables!\n");
		return NULL;
	};

	if ((M->errors = (double *) calloc (nout, sizeof (double))) == NULL)
	{
		printf ("Cannot allocate memory for error variables!\n");
		return NULL;
	}

	if ((M->outputs = (double *) calloc (nout, sizeof (double))) == NULL)
	{
		printf ("Cannot allocate memory for output vector!\n");
		return NULL;
	}  

	/* allocate memory for rule performances (MAXRULES entries) */
	if ((M->performance = (double *) calloc (MAXRULES, sizeof (double))) == NULL)
	{
		printf ("Cannot allocate memory for rule performances!\n");
		return NULL;
	}  

	/* allocate memory for sum of rule hits (MAXRULES entries) */
	if ((M->sumhits = (double *) calloc (MAXRULES, sizeof (double))) == NULL)
	{
		printf ("Cannot allocate memory for sum of rule hits!\n");
		return NULL;
	}  

	/* allocate memory for the matrix that holds mean rule targets (MAXRULES rows) */
	if ((M->meanout = (matrix *) calloc (MAXRULES, sizeof (double *))) == NULL)
	{
		printf ("Cannot allocate memory for mean target matrix!\n");
		return NULL;
	}  

	/* allocate memory for the matrix that holds rule target variances (MAXRULES rows) */
	if ((M->variance = (matrix *) calloc (MAXRULES, sizeof (double *))) == NULL)
	{
		printf ("Cannot allocate memory for variance matrix!\n");
		return NULL;
	}  

	/* allocate memory for rows of mean targets and variances */
	for (i = 0; i<MAXRULES; i++)
	{
		if ((M->meanout[i] = (double *) calloc (nout, sizeof (double))) == NULL)
		{
			printf ("Cannot allocate memory for mean rule targets matrix row!\n");
			return NULL;
		} 
		if ((M->variance[i] = (double *) calloc (nout, sizeof (double))) == NULL)
		{
			printf ("Cannot allocate memory for variance matrix row!\n");
			return NULL;
		} 
	}

	/* allocate memory for the rulebase (MAXRULES pointers to rules) */
	if ((M->rulebase = (PMRule *) calloc (MAXRULES, sizeof (PMRule))) == NULL)
	{
		printf ("Cannot allocate memory for rulebase!\n");
		return NULL;
	};

	M->in = nin;
	M->out = nout;
	M->rules = 0;
	M->Error = 0;

	switch (LP->tnorm)
	{
	case TMIN: 
		M->tnorm = minimum;
		break;
	case TPROD:
		M->tnorm = prod;
		break;
	default:
		M->tnorm = minimum;
	};

	switch (LP->snorm)
	{
	case SMAX: 
		M->snorm = maximum;
		break;
	default:
		M->snorm = maximum;
	};

	switch (LP->defuzz)
	{
	case DCOG: 
		M->defuzz = COG;
		break;
	case DMOM:
		M->defuzz = MOM;
		break;
	default:
		M->defuzz = COG;
	};

	return M;
#ifdef DEBUG_LEVEL3
	printf("Leaving Create_Mamdani_Fuzzy_System");
#endif
}; /* Create Mamdani_Fuzzy_System */


PMRule Allocate_New_Mamdani_Rule (int nin, int nout)
{
	PMRule rule;

	if ((rule = (PMRule) calloc (1, sizeof (MRule))) == NULL)
	{
		printf ("Cannot allocate memory for a new rule.\n");
		return NULL;
	};
	if ((rule->antecedent = (int *) calloc(nin, sizeof(int))) == NULL)
	{
		printf ("Cannot allocate memory for a new rule.\n");
		return NULL;
	};
	if ((rule->conclusion = (int *) calloc(nout, sizeof(int))) == NULL)
	{
		printf ("Cannot allocate memory for a new rule.\n");
		return NULL;
	};
	return rule;
};

/* This resets the hits of all output fuzzy sets of the fuzzy system to 0*/
void Reset_Mamdani_Output (MFuzzySystem *M)
{
	int i,j;

	for (i = 0; i < M->out; i++)
		for (j = 0; j < M->outvars[i]->count; j++)
			M->outvars[i]->fsets[j].hit = 0;
}; /* Reset Mamdani Output */

/* This resets the rule performances, mean rule targets and rule target variances to 0*/
void Reset_Mamdani_Statistics (MFuzzySystem *M)
{
	int i,j;

	for (i = 0; i < MAXRULES; i++)
	{
		M->performance[i] = 0;
		M->sumhits[i] = 0;
		for (j = 0; j < M->out; j++)
		{
			M->meanout[i][j] = 0;
			M->variance[i][j] = 0;
		}
	}
}; /* Reset Mamdani Statistics */

/* This computes the fulfilments of all rules of a Mamdani fuzzy system */
/* The membership degrees for all input fuzzy sets must have been computed already by */
/* calling Compute_Pattern_Memberships for the current pattern */
void Compute_Mamdani_Rule_Fulfillments (MFuzzySystem *M)
{
	int i;
	int j;
	double hit;
	PMRule rule;
	int *antec;
	int *concl;

	for (i = 0; i < M-> rules; i++)
	{
		hit = 1;
		rule = M->rulebase[i];
		antec = rule->antecedent;
		concl = rule->conclusion;

		for (j = 0; j < M->in; j++)
		{
			if (antec[j] >= 0)
				hit = M->tnorm(hit, M->invars[j]->fsets[antec[j]].hit);
		};
		rule->hit = hit;

		for (j = 0; j < M->out; j++)
		{
			if (hit > 0)
			{
				M->outvars[j]->fsets[concl[j]].hit = 
					M->snorm (hit, M->outvars[j]->fsets[concl[j]].hit);
			};
		};
	};
}; /* Compute Mamdani Rule Fulfillments */

/* This computes the fulfilments of a specified rule of a Mamdani fuzzy system */
/* The membership degrees for all input fuzzy sets must have been computed already by */
/* calling Compute_Pattern_Memberships for the current pattern */
void Compute_Mamdani_Single_Rule_Fulfillment (MFuzzySystem *M, PMRule rule)
{
	int i;
	int j;
	double hit;
	int *antec;
	int *concl;

	hit = 1;
	antec = rule->antecedent;
	concl = rule->conclusion;

	// apply t-norm to combine all antecedent hits 
	for (j = 0; j < M->in; j++)
	{
		if (antec[j] >= 0)
			hit = M->tnorm(hit, M->invars[j]->fsets[antec[j]].hit);
	};
	rule->hit = hit;

	// apply s-norm to compute consequent fulfilments
	for (j = 0; j < M->out; j++)
	{
		if (hit > 0)
		{
			if (concl[j] >= 0)
				M->outvars[j]->fsets[concl[j]].hit = 
				M->snorm (hit, M->outvars[j]->fsets[concl[j]].hit);
		};
	};
}; /* Compute Mamdani Single Rule Fulfillment */

/* This updates the statistics of a rule with the given pattern */
void Compute_Mamdani_Rule_Statistics (MFuzzySystem *M, int r, PSet *P, int which)
{
	PMRule rule;
	int i;
	double temp;

	rule = M->rulebase[r];
	Compute_Mamdani_Single_Rule_Fulfillment (M, rule);
	// if the rule fires, updates rule stats
	if (rule->hit > 0)
	{
		/* add degree of fulfilment to sum of rule hits */
		M->sumhits[r] += rule->hit; 
		/* compute the mean target sum and the mean squared target sum */
		for (i = 0; i < M->out; i++)
		{
			temp = rule->hit * P->Pats[which][P->in + i];
			M->meanout[r][i] += temp;
			M->variance[r][i] += temp * P->Pats[which][P->in + i];
		}
	}

#ifdef DEBUG_LEVEL2
	if(r==0)
	printf("Rule %i: hit=%f, target_0=%f, meanout_0=%f, sumhits=%f\n",
		r, rule->hit, P->Pats[which][P->in], M->meanout[r][0], M->sumhits[r]);
#endif
}/* Compute Mamdani Rule Statistics*/

/*
	This computes the rule statistics for all rules for the given pattern.
*/
void Compute_Mamdani_Rulebase_Statistics(MFuzzySystem *M, PSet *P, int which)
{
	int i;

	for (i = 0; i < M->rules; i++)
	{
		Compute_Mamdani_Rule_Statistics (M, i, P, which);
	}
}// Compute Mamdani Rulebase Statistics

/* Finalises the computation of all rule statistics and computes rule performances */
/* The pattern set is required to get the information about the ranges of output variables*/
void Compute_Mamdani_Rule_Performances (MFuzzySystem *M, PSet *P)
{
	int i, j;
	double range[MAXOUT];

	for (j = 0; j < M->out; j++)
	{
		range[j] = P->pmax[P->in+j] - P->pmin[P->in+j];
		if (range[j] <= 0) range[j] = 1;
	}

	for (i = 0; i < M->rules; i++)
	{
		for (j =0; j < M->out; j++)
		{
			if (M->sumhits[i] > 0) // only for rules that have ever fired
			{
				M->variance[i][j] = (M->variance[i][j] - (sqr(M->meanout[i][j]) / M->sumhits[i])) / M->sumhits[i];
				M->meanout[i][j] = M->meanout[i][j] / M->sumhits[i];
				M->performance[i] += (M->sumhits[i] * range[j]) / (P->count * (2*sqrt(M->variance[i][j]) + range[j]));
			}
		}
		if (M->performance[i] < 0) M->performance = 0;
#ifdef DEBUG_LEVEL1
		printf("Rule %i: sumhits=%f, meanout_0=%f, variance_0=%f, range_0=%f, performance=%f\n",
			i, M->sumhits[i], M->meanout[i][0], M->variance[i][0], range[0], M->performance[i]);
#endif
	}
}// Compute Mamdani Rule Performances


/* compute the final output values by defuzzification */
void Compute_Mamdani_Output (MFuzzySystem *M)
{
	int i;

	for (i = 0; i < M->out; i++)
	{
		M->outputs[i] = M->defuzz (M->outvars[i], M->tnorm, M->snorm);
	};
}; /* Compute Mamdani Output */

double Compute_Mamdani_Error (MFuzzySystem *M, PSet *P, int which)
{
	int i;
	double error;

	error = 0;
	for (i = 0; i < P->out; i++)
	{
		M->errors[i] = P->Pats[which][P->in + i] - M->outputs[i];
		error += sqr(M->errors[i]);
	};
	return (error);

};/* Compute Mamdani Error */

double Mamdani_Propagate_Pattern (MFuzzySystem *M, PSet *P, int which)
{
	double error;

	Compute_Pattern_Memberships (M->invars, P->Pats[which], 0, P->in);

	Reset_Mamdani_Output (M);

	Compute_Mamdani_Rule_Fulfillments (M);

	Compute_Mamdani_Output (M);

	error = Compute_Mamdani_Error (M, P, which);
	M->Error += error;
	return error;

}; /* Mamdani Propagate Pattern */


/* 
   This creates a new rulebase for a Mamdani fuzzy system or adds rules to an 
   an existing rule base if it already exists.
   First the rule antecedents will be identified by finding for each pattern a
   combination of fuzzy sets that each give the largest degree of membership for
   the current pattern. If this combination does not yet exist in a rule, a new
   rule is added, if the maximum number of rules has not been reached, yet.
   After all patterns have been processed, the optimal consequent for each 
   new rule is identified by using the weighted average of all pattern targets
   for each rule output. For this value the fuzzy set that provides the 
   largest degree of membership is used.
   The required statistical values for that (mean output, variance, 
   sum of hits) are kept in the Mamdani Fuzzy System object.
   This algorithm is documented on page 60 of
   http://fuzzy.cs.uni-magdeburg.de/~nauck/nauck.pdf.
   The only difference is that it does not support the on-the-fly creation
   of new output fuzzy sets, but uses pre-defined partitions.
*/
int Create_Mamdani_Rulebase (MFuzzySystem *M, PSet *P)
{
	int i,j,k,l;
	int imaxfs[MAXIN];
	int omaxfs[MAXOUT];
	int index;
	int equal;
	double max;
	double hit;
	double error;
	int start;

	start = M->rules;

	for (i = 0; i < P->count; i++)
	{ /* process all patterns */
		Compute_Pattern_Memberships (M->invars, P->Pats[i], 0, P->in);

		for (j = 0; j < M->in; j++)
		{ /* find best antecedent for the current pattern */
			max = 0;
			index = -1;
			for (k = 0; k < M->invars[j]->count; k++)
			{
				if (max < (hit = M->invars[j]->fsets[k].hit))
				{
					max = hit;
					index = k;
				};
			}; /* for k */
			if (index < 0)
			{
				printf ("Error in fuzzy partition of input %i.\n", j);
				return 0;
			}
			else 
				imaxfs[j] = index;
		}; /* for j */

		equal = 0; j = 0;
		while (!equal && (j < M->rules))
		{ /* check, whether there is already such an antecedent */
			k = 0; equal = 1;
			while (equal && (k < M->in))
			{
				equal &= (imaxfs[k] == M->rulebase[j]->antecedent[k]);
				k++;
			};
			// found the antecedent in the rule base: rule j
			if (equal)
			{
				// compute the statistics for rule j and current pattern i
				Compute_Mamdani_Rule_Statistics (M, j, P, i);
			};
			j++;
		}; /* while (!equal) */

		if ((!equal) && (M->rules < MAXRULES))
		{ 
			/* there is no such antecedent yet, so lets create a new rule, */
			/* if we are still allowed to do this;   */

			/* allocate a new rule, and set antecedent and conclusion */
			M->rulebase[M->rules] = Allocate_New_Mamdani_Rule(M->in, M->out);
			for (j = 0; j < M->in; j++)
			{
				M->rulebase[M->rules]->antecedent[j] = imaxfs[j];
			};
			for (j = 0; j < M->out; j++)
			{
				M->rulebase[M->rules]->conclusion[j] = -1; // no consequent at this stage
			};
			Compute_Mamdani_Rule_Statistics (M, M->rules, P, i);
			M->rules++;
		}; /* if (!equal) */

	}; /* for i (end of pattern loop) */

	/* Now compute statistics of the (still incomplete) rules */
	Compute_Mamdani_Rule_Performances(M, P);

	/* Now find best consequent fuzzy sets for the rules, */
	/* but only for new ones */
	for (i = start; i < M->rules; i++)
	{
		/* get degrees of fulfilment for average rule targets for output fuzzy sets */	
		Compute_Pattern_Memberships (M->outvars, M->meanout[i], 0, M->out);
		for (j = 0; j < M->out; j++)
		{
			max = 0; 
			index = -1;
			/* for all output variables find fuzzy set with largest degree of membership */
			for (k = 0; k < M->outvars[j]->count; k++)
			{
				if (max < (hit = M->outvars[j]->fsets[k].hit))
				{
					max = hit;
					index = k;
				};
			}; /* for k */
			if (index < 0)
			{
				printf ("Error in fuzzy partition of output %i.\n",j);
				return 0;
			}
			else 
			{
				/* for a new rule this will always be true */
				if (index != M->rulebase[i]->conclusion[j])
				{
					printf ("Best consequent for rule %i, output %i: %i\n",
						i,j,index);
					M->rulebase[i]->conclusion[j] = index;
				};
			};
		}; /* for j */      
	};

	return M->rules;
}; /* Create Mamdani Rulebase*/

/*
	This evaluates all rules of a Mamdani Systems and selects the best rules 
	based on the rule performance. The learning parameters determine, if 
	a fixed number of rules is used or if a certain percentage of patterns
	have to be covered.
*/
int Evaluate_Mamdani_Rulebase (MFuzzySystem *M, PSet *P, LParam *LP)
{
	int i,j;
	int stop;
	int next;
	int rules;
	int maxrules;
	double error;
	double maximum;
	double sumhit;
	double hits;
	PMRule *newrulebase; 
	PMRule *oldrulebase;
	FILE *rulefile;

	/* recompute the rule performance, in case they are not there, yet */
	Reset_Mamdani_Statistics(M);
	for (i = 0; i < P->count; i++)
	{
		error = Mamdani_Propagate_Pattern (M, P, i);
		Compute_Mamdani_Rulebase_Statistics (M, P, i);
	};
	Compute_Mamdani_Rule_Performances(M, P);

	rulefile = fopen ("rules.txt", "w");
	fprintf (rulefile, "Rule\tPerformance\tMean_Target_0\tVariance_0\tMean_Hit\n");
	sumhit = 0;
	for (i = 0; i < M->rules; i++)
	{
		fprintf (rulefile, "Rule %i:\t%f\t%f\t%f\t%f\n", i, M->performance[i], M->meanout[i][0], 
			M->variance[i][0], M->sumhits[i]/P->count);
		sumhit += M->sumhits[i];
	};

	fprintf (rulefile, "Sum of hits: %f\n\nNew Rules\n", sumhit);

	/* allocate memory for the new rulebase (maxrules pointers to rules) */
	if ((M->rules < LP->maxrules) || (LP->maxrules <= 0))
		maxrules = M->rules;
	else
		maxrules = LP->maxrules;

	if ((newrulebase = (PMRule *) calloc (maxrules, sizeof (PMRule))) == NULL)
	{
		printf ("Cannot allocate memory for new rulebase!\n");
		exit (-1);
	};

	i = 0; stop = 0; hits = 0;

	while (!stop)
	{
		maximum = -1;
		next = 0;
		for (j = 0; j < M->rules; j++)
		{
			if ((M->performance[j] > maximum) && (M->performance[j] >= 0))
			{
				maximum = M->performance[j];
				next = j;
			};
		}; /* for j */

		fprintf (rulefile, "New rule %i is former rule %i: %f\n", 
			i, next, M->performance[next]);

		/* allocate a new rule, and set antecedent and conclusion */
		if ((newrulebase[i] = Allocate_New_Mamdani_Rule(M->in, M->out)) 
			== NULL)
		{
			printf ("Cannot allocate new rule for new rulebase.\n");
			exit (-1);
		};
		for (j = 0; j < M->in; j++)
		{
			newrulebase[i]->antecedent[j] = 
				M->rulebase[next]->antecedent[j];
		};
		for (j = 0; j < M->out; j++)
		{
			newrulebase[i]->conclusion[j] = 
				M->rulebase[next]->conclusion[j];
		};

		M->performance[next] = -1;
		i++;
		hits += M->sumhits[next];

		if (LP->maxrules <= 0)
		{
			stop = ((hits >= LP->percent * sumhit) || (i >= maxrules));
		}
		else
		{
			stop = (i >= maxrules);
		};
	}; /* while */

	oldrulebase = M->rulebase;
	rules = M->rules;
	M->rulebase = newrulebase;
	M->rules = i;

	for (i = 0; i < rules; i++)
	{
		free (oldrulebase[i]->antecedent);
		free (oldrulebase[i]->conclusion);
		free (oldrulebase[i]);
	};
	free (oldrulebase);
	fclose (rulefile);
	return M->rules;

}; /* Evaluate Mamdani Rulebase */

int Mamdani_Test_Pattern (MFuzzySystem *M, PSet *P, const char *filename)
{
	int i,j;
	FILE *resultfile;


	if ((resultfile = fopen (filename, "w")) == NULL)
	{
		printf ("Cannot open result file.\n");
		return -1;
	};

	for (i = 0; i < P->count; i++)
	{
		Mamdani_Propagate_Pattern (M, P, i);
		for (j = 0; j < P->out; j ++)
		{
			fprintf (resultfile, "%10.4f %10.4f %10.4f", P->Pats[i][P->in+j],
				M->outputs[j],  P->Pats[i][P->in+j] - M->outputs[j]);
		};
		fprintf (resultfile, "\n");
	};
	fclose (resultfile);
	return 0;
}; /* Mamdani_Test_Pattern */


void CopyPartition (FSet *F, FSet *Copy, int count)
{
	int i;
	for (i = 0; i < count; i++)
	{
		Copy[i].a = F[i].a;
		Copy[i].b = F[i].b;      
		Copy[i].c = F[i].c;
		Copy[i].d = F[i].d;
		Copy[i].ls = F[i].ls;
		Copy[i].rs = F[i].rs;
		Copy[i].form = F[i].form;
		strcpy (Copy[i].name, F[i].name);
	};
};

void CopyLV (LingVar *LV, LingVar *Copy)
{


	strcpy (Copy->name, LV->name);
	Copy->upper = LV->upper;
	Copy->lower = LV->lower;
	Copy->count = LV->count;
	Copy->form  = LV->form;
	CopyPartition (LV->fsets, Copy->fsets, LV->count);
};

void Copy_Mamdani (MFuzzySystem *M, MFuzzySystem **Copy, LParam *LP)
{
	int i, j;
	MFuzzySystem *MC;

	if (M != NULL)
	{
		MC = *Copy;
		if (MC == NULL)
		{
			MC = Create_Mamdani_Fuzzy_System (M->in, M->out, LP);

			for (i = 0; i < M->in; i++)
			{
				if ((MC->invars[i] = 
					(LingVar *) calloc (1, sizeof(LingVar))) == NULL)
				{
					printf ("Failed to copy %ith input variable.\n", i);
					exit (-(i+1));
				};
				CopyLV (M->invars[i], MC->invars[i]);
			};

			for (i = 0; i < M->out; i++)
			{
				if ((MC->outvars[i] = 
					(LingVar *) calloc (1, sizeof(LingVar))) == NULL)
				{
					printf ("Failed to copy %ith output variable.\n", i);
					exit (-(i+1));
				};
				CopyLV (M->outvars[i], MC->outvars[i]);
			};	  

			for (i = 0; i < M->rules; i++)
			{
				MC->rulebase[i] = 
					Allocate_New_Mamdani_Rule(MC->in, MC->out);
				MC->rules++;
				for (j = 0; j < M->in; j++)
				{
					MC->rulebase[i]->antecedent[j] = 
						M->rulebase[i]->antecedent[j];
				};
				for (j = 0; j < M->out; j++)
				{
					MC->rulebase[i]->conclusion[j] = 
						M->rulebase[i]->conclusion[j];
				};
			};
			*Copy = MC;
		}/* creation of new MFuzzySystem completed */
		else 
		{ /* there is already a copy, just update MSF parameters */
			for (i = 0; i < M->in; i++)
			{
				for (j = 0; j < M->invars[i]->count; j++)
				{
					MC->invars[i]->fsets[j].a = M->invars[i]->fsets[j].a;
					MC->invars[i]->fsets[j].b = M->invars[i]->fsets[j].b;      
					MC->invars[i]->fsets[j].c = M->invars[i]->fsets[j].c;
					MC->invars[i]->fsets[j].d = M->invars[i]->fsets[j].d;
				};
			};
			for (i = 0; i < M->out; i++)
			{
				for (j = 0; j < M->outvars[i]->count; j++)
				{
					MC->outvars[i]->fsets[j].a = M->outvars[i]->fsets[j].a;
					MC->outvars[i]->fsets[j].b = M->outvars[i]->fsets[j].b;
					MC->outvars[i]->fsets[j].c = M->outvars[i]->fsets[j].c;
					MC->outvars[i]->fsets[j].d = M->outvars[i]->fsets[j].d;
				};
			};
		}; /* update of existing copy completed */
	};
} /* Copy_Mamdani*/

