/* NFIDENT: a program for neuro-fuzzy function approximation

   Copyright (C) 1997-2000 

   Detlef Nauck (nauck@ieee.org, 
   http:\\fuzzy.cs.uni-magdeburg.de/~nauck)  
   hereby disclaims all copyright interests in the program 
   NFIDENT written by Detlef Nauck

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 1, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/* nefprox.c */
/* This file contains functions to train a Mamdani fuzzy system by the */
/* NEFPROX learning algorithm */

#include <math.h>
#include <stdio.h>
#include "global.h"
#include "fuzzy.h"
#include "mamdani.h"

/* Reset_Error is a local function to reset all error and delta values */
/* of a Mamdani fuzzy system. */
/* Parameters: M: pointer to a Mamdani fuzzy system */
void Nefprox_Reset_Delta (MFuzzySystem *M)
{
  int i,j;
  FSet *F;

  /* reset delta values of input fuzzy sets */
  for (i = 0; i < M->in; i++)
    {
      F = M->invars[i]->fsets;
      for (j = 0; j < M->invars[i]->count; j++)
	{
	  F[j].d_a = 0; F[j].d_b = 0; F[j].d_c = 0; F[j].d_d = 0;
	};
    };
  /* reset delta values of output fuzzy sets */
  for (i = 0; i < M->out; i++)
    {
      F = M->outvars[i]->fsets;
      for (j = 0; j < M->outvars[i]->count; j++)
	{
	  F[j].d_a = 0; F[j].d_b = 0; F[j].d_c = 0; F[j].d_d = 0;
	};
    };

}; /* Reset Delta */

void Nefprox_Compute_Delta (MFuzzySystem *M, double *Pat, LParam *LP)
{
  int i, j, k;
  FSet *F;
  double woutdelta; /* weighted output delta of an output variable */
  double x;         /* supporting variables */
  double d_b;
  PMRule rule;

  Compute_Pattern_Memberships (M->outvars, Pat, M->in, M->out);
  for (i = 0; i < M->rules; i++)
    {
      if (M->rulebase[i]->hit > 0)
	{
	  rule = M->rulebase[i];
	  rule->error = 0;
	  for (j = 0; j < M->out; j ++)
	    {
	      F = M->outvars[j]->fsets;
	      k = rule->conclusion[j];
	      /* woutdelta reflects, whether the influence of the current   */
	      /* output fuzzy set should be increased (membership of target */
	      /* > 0.5) or decreased (membership of target < 0.5).          */
	      woutdelta = (2.0 * F[k].hit - 1.0) *  fabs (M->errors[j]);

	      rule->error += woutdelta;

	      /* Idea for the modifications of the output fuzzy sets:     */
	      /* If the target value has a high membership degree, then   */
	      /* shift the fuzzy set a little bit to the target, and make */
	      /* it narrower. If the membership degree is zero, then shift*/
	      /* it to the target, and make it wider on the side of       */
	      /* target, and narrower on the opposite side.               */
	      /* Considerations for future work:                          */
	      /* the current version seems to work well for MOM, what     */
	      /* about COG?                                               */
	      /* The delta values are accumulated (+=) because the fuzzy  */
	      /* set are shared by the rules.                             */
	      switch (F[k].form)
		{
		case TRIANGLE:
		  woutdelta *= (F[k].c - F[k].a) * rule->hit;
		  d_b = LP->sigma_b * M->errors[j] * (F[k].c - F[k].a) * 
		    (1.0 - F[k].hit) * rule->hit;
		  /*		  if (!LP->asym || 
		      ((Pat[M->in + j] <= F[k].b) && (F[k].hit > 0.5)) ||
		      ((Pat[M->in + j] >= F[k].b) && (F[k].hit < 0.5)))
		    {
		      F[k].d_a += - LP->sigma_a * woutdelta + F[k].d_b;
		    };
		  if (!LP->asym || 
		      ((Pat[M->in + j] >= F[k].b) && (F[k].hit > 0.5)) ||
		      ((Pat[M->in + j] <= F[k].b) && (F[k].hit < 0.5)))
		    {
		      F[k].d_c += LP->sigma_c * woutdelta + F[k].d_b;
		    };
		    */
		  if (F[k].hit > 0)
		    {
		      F[k].d_a += LP->sigma_a * woutdelta + d_b;
		      F[k].d_c += - LP->sigma_c * woutdelta + d_b;
		      F[k].d_b += d_b;
		    }
		  else 
		    {
		      F[k].d_b += d_b;
		      if (LP->asym)
			{
			  if (Pat[M->in + j] > F[k].b)
			    {
			      F[k].d_a +=  LP->sigma_a * woutdelta + d_b;
			      F[k].d_c +=  LP->sigma_c * woutdelta + d_b;
			    }
			  else
			    {
			      F[k].d_a +=  - LP->sigma_a * woutdelta + d_b;
			      F[k].d_c +=  - LP->sigma_c * woutdelta + d_b;
			    };
			}
		      else /* if asym=0, just make the fuzzy set narrower */
			{
			  F[k].d_a +=  LP->sigma_a * woutdelta + d_b;
			  F[k].d_c +=  - LP->sigma_c * woutdelta + d_b;
			};
		    };
		  break;
		case TRAPEZOID:
		  break;
		case GAUSSIAN:
		  woutdelta *= F[k].a;
		  d_b = LP->sigma_b * M->errors[j] * F[k].a * 
		        (1.0 - F[k].hit) * rule->hit;
		  F[k].d_b += d_b;
		  F[k].d_a = 0;
		  /* if (F[k].hit > 0)
		    { 
		      F[k].d_a +=  LP->sigma_a * woutdelta;
		    }
		   else 
		    { 
		      F[k].d_a -=  LP->sigma_a * woutdelta;
		    }; */
		  break;
		}; /* switch */
	    }; /* for j*/
	  /* the rule error reflects, whether the influence of the rule  */
	  /* must be increased (error > 0) or decreased (error < 0). The */
	  /* rule error is larger for a rule fulfillment near 0.5.       */
	  rule->error *= 
	    rule->hit * (1.0 - rule->hit);

	  /* Determine the delta values for the input fuzzy sets. Depending */
	  /* on the rule error, a fuzzy set is moved towards the current    */
	  /* input, and is made wider (results in higher membership degree),*/
	  /* or is shifted away from the input, and is made narrower.       */
	  /* A fuzzy set with a smaller degree of membership is changed more*/
	  /* strongly than a fuzzy set with a higher degree of membership.  */
	  for (j = 0; j < M->in; j++)
	    {
	      F = M->invars[j]->fsets;
	      k = rule->antecedent[j];
	      switch (F[k].form)
		{
		case TRIANGLE:
		  x = rule->error * (F[k].c - F[k].a) 
		    * (1.0 - F[k].hit);
		  d_b = LP->sigma_b * x * sgn (Pat[j] - F[k].b);
		  if (!LP->asym || (Pat[j] <= F[k].b))
		    {
		      F[k].d_a += - LP->sigma_a * x + d_b;
		    };
		  if (!LP->asym || (Pat[j] >= F[k].b))
		    {
		      F[k].d_c += LP->sigma_c * x + d_b;
		    };
		  F[k].d_b += d_b;
		  break;
		case TRAPEZOID:
		  break;
		case GAUSSIAN:
		  x = rule->error * F[k].a * (1.0 - F[k].hit);
		  d_b = LP->sigma_b * x * sgn (Pat[j] - F[k].b);
		  F[k].d_a +=  LP->sigma_a * x;
		  F[k].d_b += d_b;
		  break;
		}; /* switch */
	    }; /* for j*/
	}; /* if */
    }; /* for i*/
}; /* Compute Delta */


/* Adapt carries out the apaptions of the membership functions, by calling  */
/* Adapt_LingVar for each linguistic Variable. norm is used to normalize    */
/* the modifications (1 for online mode, number of patterns for batch mode) */


void Nefprox_Adapt (MFuzzySystem *M, LParam *LP, int norm)
{
  int i;

  /* adapt input membership functions (learning rate already applied) */
  for (i = 0; i < M->in; i++)
    {
      Adapt_LingVar (M->invars[i], norm, LP->nopass, LP->at05);
    }; /* for i*/

  /* adapt output membership functions (learning rate already applied) */
  for (i = 0; i < M->out; i++)
    {
      Adapt_LingVar (M->outvars[i], norm, 0, 0);
    }; /* for i*/
}; /* Adapt */

/* Nefprox_Training is the main training loop for NEFPROX. Training stops,  */
/* if a) the desired error is reached, or b) the maximum number of epochs */
/* has passed, or c) the error has not improved for a given time. In any  */
/* case, training is carried out for at least a minmum number of epochs.  */
/* Parameters: M: pointer to Mamdani fuzzy system */
/*             P: pointer to training patterns */
/*             LP: pointer to learning parameters */
/*             G: pointer to generalization patterns (can be NULL) */
/*             return result: number of training epochs */

int Nefprox_Training (MFuzzySystem *M, PSet *P, LParam *LP, PSet *G)
{
  int i;
  int stop;
  int epoch;
  int history;
  int last_change;
  int last_improve;
  double best_error;
  double train_error;
  double last_error;
  MFuzzySystem *Save;
  FILE *logfile;
  char savefile[80];

  stop = 0;
  epoch = 0; 
  history = 0;
  last_change = 0;
  last_improve = 0;
  last_error = 0;
  Save = NULL;

  if ((logfile = fopen (LP->logfilename, "w")) == NULL)
    {
      printf ("Cannot create logfile %s!\n", LP->logfilename);
      return -1;
    };
  
  Nefprox_Reset_Delta(M);

  while (!stop)
    {
      epoch++;
      M->Error = 0; 
      Copy_Mamdani (M, &Save, LP);
      /* propagte training patterns, and determine error and delta values */
      for (i = 0; i < P->count; i++)
	{
	  Mamdani_Propagate_Pattern (M, P, i);
	  Nefprox_Compute_Delta (M, P->Pats[i], LP);
	  /* do a learning step, if not in batch mode */
	  if (!LP->batch)
	    {
	      Nefprox_Adapt (M, LP, 1);
	      Nefprox_Reset_Delta (M);
	    };
	};
      M->Error = sqrt (M->Error / (P->count * M->out));
      /* if in batch mode, do a learning step now */
      if (LP->batch)
	{
	  Nefprox_Adapt (M, LP, P->count);
	  Nefprox_Reset_Delta (M);
	};

      train_error = M->Error;

      if (G != NULL)
	{ /* if generalization patterns are given, do cross-validation */
	  M->Error = 0;
	  for (i = 0; i < G->count; i++)
	    {
	      Mamdani_Propagate_Pattern (M, G, i);
	    };
	  M->Error = sqrt (M->Error / (G->count * M->out));
	};
 
      /* store best error value, and the epoch it occured in, and save net */
      if (epoch == 1)
	{
	  best_error = M->Error;
	  last_error = M->Error;
	  last_improve = 1;
	  Copy_Mamdani (Save, &BestNefprox, LP);
	}
      else
	{
	  if (best_error > M->Error)
	    {
	      best_error = M->Error;
	      last_improve = epoch;
	      if (G != NULL)
		Copy_Mamdani (M, &BestNefprox, LP);
	      else
		Copy_Mamdani (Save, &BestNefprox, LP);
	    };
	  if (last_error <  M->Error)
	    {
	      history <<=1;
	      history++;
	    }
	  else
	    {
	      history <<=1;
	    };
	  
	  if (epoch - last_change >= 4)
	    {
	      history &= 0xF;
	      
	      if (history == 0)
		{
		  LP->sigma_a *= 1.1;
		  LP->sigma_b *= 1.1;
		  LP->sigma_c *= 1.1;
		  LP->sigma_d *= 1.1;
		  LP->sigma_t *= 1.1;
		  printf ("Learning rate increased in epoch %i, sigma_a = %f (%i).\n", 
			  epoch, LP->sigma_a, history);
		};
	      if ((history == 3) || (history == 5) || (history == 7) ||
		  (history == 11)|| (history == 15)) 
		{
		  LP->sigma_a *= 0.9;
		  LP->sigma_b *= 0.9;
		  LP->sigma_c *= 0.9;
		  LP->sigma_d *= 0.9;
		  LP->sigma_t *= 0.9;
		  printf ("Learning rate decreased in epoch %i, sigma_a = %f (%i).\n", 
			  epoch, LP->sigma_a, history);
		};

	      last_change = epoch;
	    };
	};
      last_error = M->Error;

      /* check, whether we must stop now */
      stop = (((M->Error <= LP->minerror) || (epoch > LP->maxepoch) ||
	      (epoch - last_improve > LP->chgepoch)) && 
	     (epoch > LP->minepoch)) || StopOnSignal;
      if (((epoch % LP->logepoch) == 0) || (epoch == 1) || stop)
	{
	  if (G != NULL)
	    {
	      printf ("Epoch %i: training error: %f, ", epoch, train_error);
	      printf ("generalization error: %f,\n", M->Error);
	      printf ("best error of %f in epoch %i.\n", 
		      best_error, last_improve);
	    }
	  else
	    {
	      printf ("Epoch %i: Error = %f, best error of %f in epoch %i.\n",
		  epoch, M->Error, best_error, last_improve);
	    };
	  fprintf (logfile, "%i %f %f %i\n", 
		   epoch, M->Error, best_error, last_improve);
	  strcpy (savefile, LP->savenetfilename);
	  Build_Filename (savefile, epoch);
	  Write_MFuzzy_System (savefile, BestNefprox);
	};
    };
  fclose (logfile);
  return epoch;
};/* Nefprox Training */


