/* NFIDENT: a program for neuro-fuzzy function approximation

   Copyright (C) 1997-2000 

   Detlef Nauck (detlef.nauck@cs.uni-magdeburg.de, 
   http:\\fuzzy.cs.uni-magdeburg.de/~nauck)  
   hereby disclaims all copyright interests in the program 
   NFIDENT written by Detlef Nauck

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 1, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <string.h>
#include <math.h>
#include <signal.h>
#include "global.h"
#include "pattern.h"
#include "fuzzy.h"
#include "mamdani.h"
#include "defuzz.h"
#include "tsk.h"
#include "interfac.h"
#include "nefprox.h"
#include "anfis.h"


int Mamdani_Training_Cycle()
{
  int i, j;
  double x = 0.0;
  char mode[3];
  double Error = 0.0;
  double *errors;
  int epochs;  

  /* if there is no Nefprox, then create one */
  if (Nefprox == NULL)
    {
      printf ("Creating Nefprox.\n");
      if ((Nefprox = Create_Mamdani_Fuzzy_System (TrainingPattern->in, 
						 TrainingPattern->out, 
						 &LP)) 
	  == NULL)
	{
	  printf ("Failed to create Nefprox system, giving up.\n");
	  return -1;
	};
      
      printf("Creating Linguistic Variables.\n");
      if (Create_Linguistic_Variables (Nefprox->invars, Nefprox->outvars, 
				       TrainingPattern, 
				       GeneralizationPattern,
				       TestPattern, &LP, MAMDANI)
	  < 0)
	{
	  printf ("Failed to create linguistic variables, giving up.\n");
	  return -1;
	};
    };
  
  /* allocate memory for output errors */
  if ((errors = (double *) calloc (TrainingPattern->out, sizeof (double))) 
      == NULL)
    {
      printf ("Cannot allocate memory for errors.\n");
      exit (-1);
    };

  /* create a rulebase for Nefprox, or try to add rules if there is already a */
  /* rulebase */
  printf ("Now creating a rule base.\n");
  i = Create_Mamdani_Rulebase (Nefprox, TrainingPattern);
  printf ("Created %i rules.\n", i);
  
  /* evaluate the rulebase */
  if (LP.maxrules >= 0)
    {
      printf ("Now evaluating the rule base.\n");
      i = Evaluate_Mamdani_Rulebase (Nefprox, TrainingPattern, &LP);
      printf ("Selected best %i rules.\n", i);
    };
  
  /* save the fuzzy system before training the membership functions */
  printf ("Write fuzzy system to before.nfs.\n\n");
  Write_MFuzzy_System ("before.nfs", Nefprox);
  
  /* training */
  printf ("Now starting the training process for %i epochs.\n",
	  LP.maxepoch);
  printf ("Press Ctrl-C to stop.\n"); 
  printf ("====================================================\n\n");

  epochs = Nefprox_Training (Nefprox, TrainingPattern, &LP, 
			    GeneralizationPattern);
  printf ("Training stopped after %i epochs.\n", epochs);
  
  /* restore and save the best fuzzy system */
  printf ("Restoring best fuzzy system.\n");
  Copy_Mamdani (BestNefprox, &Nefprox, &LP);
  printf ("Write best fuzzy system to %s fuzzy system.\n", LP.netfilename);
  Write_MFuzzy_System (LP.netfilename, Nefprox);
  
  /* write the fuzzy partitions to a gnuplot command file */
  printf ("Write gnuplot command file %s to plot the fuzzy sets.\n",
	  LP.gnufilename);
  strcpy (mode, "w");
  for (i = 0; i < TrainingPattern->in; i++)
    {
      Write_Partition_For_Gnuplot (LP.gnufilename, Nefprox->invars[i], mode); 
      strcpy (mode, "a");
    };
  
  for (i = 0; i < TrainingPattern->out; i++)
    {
      Write_Partition_For_Gnuplot (LP.gnufilename, Nefprox->outvars[i], "a");
    };
  

  /* Test all pattern sets */
  
  /* Training patterns */
  Error = 0; Nefprox->Error = 0;
  for (j = 0 ; j< Nefprox->out; j++)
    errors[j] = 0;
  printf ("Now propagating training patterns.\n");
  for (i = 0; i < TrainingPattern->count; i++)
    {
      x = Mamdani_Propagate_Pattern (Nefprox, TrainingPattern, i);
      Error += x;
      for (j = 0 ; j< Nefprox->out; j++)
	errors[j] += sqr(Nefprox->errors[j]);
    };
  
  printf ("Overall Root Mean Square Error: %f.\n", 
	  sqrt (Error / (TrainingPattern->count * Nefprox->out)));
  for (j = 0; j < Nefprox->out; j++)
    {
      printf ("Output %i, Non-Dimensional Error Index: %f.\n", j,
	      sqrt(errors[j]/TrainingPattern->count) / 
	      TrainingPattern->pstddev[Nefprox->in + j]);
    };

  printf ("Write training result file %s.\n", LP.trainresultname);
  Mamdani_Test_Pattern (Nefprox, TrainingPattern, LP.trainresultname);
    
  /* Generalization patterns */
  if (GeneralizationPattern != NULL)
    {
      Error = 0; Nefprox->Error = 0;
      for (j = 0 ; j< Nefprox->out; j++)
	errors[j] = 0;
      printf ("Now propagating generalization patterns.\n");
      for (i = 0; i < GeneralizationPattern->count; i++)
	{
	  x = Mamdani_Propagate_Pattern (Nefprox, GeneralizationPattern, i);
	  Error += x;
	  for (j = 0 ; j< Nefprox->out; j++)
	    errors[j] += sqr(Nefprox->errors[j]);
	};
      printf ("Overall Root Mean Square Error: %f.\n", 
	      sqrt (Error / (GeneralizationPattern->count * Nefprox->out)));
      for (j = 0; j < Nefprox->out; j++)
	{
	  printf ("Output %i, Non-Dimensional Error Index: %f.\n", j,
		  (sqrt(errors[j]/GeneralizationPattern->count) / 
		   GeneralizationPattern->pstddev[Nefprox->in + j]));
	};
      
      printf ("Write generalization result file %s.\n", LP.genresultname);
      Mamdani_Test_Pattern (Nefprox, GeneralizationPattern, LP.genresultname);
    };

  /* Test patterns */
  if (TestPattern != NULL)
    {
      Error = 0; Nefprox->Error = 0;
      for (j = 0 ; j< Nefprox->out; j++)
	errors[j] = 0;
      printf ("Now propagating test patterns.\n");
      for (i = 0; i < TestPattern->count; i++)
	{
	  x = Mamdani_Propagate_Pattern (Nefprox, TestPattern, i);
	  Error += x;
	  for (j = 0 ; j< Nefprox->out; j++)
	    errors[j] += sqr(Nefprox->errors[j]);
	};
      
      printf ("Overall Root Mean Square Error: %f.\n", 
	      sqrt (Error / (TestPattern->count * Nefprox->out)));
      for (j = 0; j < Nefprox->out; j++)
	{
	  errors[j] = sqrt(errors[j] / TestPattern->count) / 
	    TestPattern->pstddev[Nefprox->in + j];
	  
	  printf ("Output %i, Non-Dimensional Error Index: %f.\n", 
		  j, errors[j]);
	};
      printf ("Write test result file %s.\n", LP.testresultname);
      Mamdani_Test_Pattern (Nefprox, TestPattern, LP.testresultname);
    };

  return epochs;

}; /* Mamdani Training Cycle */

/**************************************************************************/

int TSK_Training_Cycle ()
{
  int i, j;
  double x = 0.0;
  char mode[3];
  double Error = 0.0;
  double *errors;
  int epochs;  

  /* if there is no Anfis, then create one */
  if (Anfis == NULL)
    {
      printf ("Creating ANFIS.\n");
      if ((Anfis = Create_TSK_Fuzzy_System (TrainingPattern->in, 
					    TrainingPattern->out, 
					    &LP)) 
	  == NULL)
	{
	  printf ("Failed to create ANFIS system, giving up.\n");
	  return -1;
	};
      
      printf("Creating Linguistic Variables.\n");
      if (Create_Linguistic_Variables (Anfis->invars, NULL,
				       TrainingPattern,  
				       GeneralizationPattern,
				       TestPattern, &LP, SUGENO)
	  < 0)
	{
	  printf ("Failed to create linguistic variables, giving up.\n");
	  return -1;
	};
    };

  /* allocate memory for output errors */
  if ((errors = (double *) calloc (TrainingPattern->out, sizeof (double))) 
      == NULL)
    {
      printf ("Cannot allocate memory for errors.\n");
      exit (-1);
    };


  /* create a rule base for Anfis, or try to add rules to an existing one */
  printf ("Now creating a rule base.\n");
  i = Create_TSK_Rulebase (Anfis, TrainingPattern, &LP);
  printf ("Created %i rules.\n", i);
  
  
  /* save the fuzzy system before training the membership functions */
  printf ("Write fuzzy system to before.nfs.\n\n");
  Write_TSKFuzzy_System ("before.nfs", Anfis);
  
  /* Training */
  printf ("Now starting the training process for %i epochs.\n", 
	  LP.maxepoch);
  printf ("Press Ctrl-C to stop.\n"); 
  printf ("====================================================\n\n");

  epochs = Anfis_Training (Anfis, TrainingPattern, &LP, 
			   GeneralizationPattern);
  printf ("Training stopped after %i epochs.\n", epochs);
  
  /* restore and save the best fuzzy system */
  printf ("Restoring best fuzzy system.\n");
  Copy_TSK (BestAnfis, &Anfis, &LP);
  printf ("Write best fuzzy system to %s fuzzy system.\n", LP.netfilename);
  Write_TSKFuzzy_System (LP.netfilename, Anfis);
  
  /* write the fuzzy partitions to a gnuplot command file */
  printf ("Write gnuplot command file %s to plot the fuzzy sets.\n",
	  LP.gnufilename);
  strcpy (mode, "w");
  for (i = 0; i < TrainingPattern->in; i++)
    {
      Write_Partition_For_Gnuplot (LP.gnufilename, Anfis->invars[i], mode); 
      strcpy (mode, "a");
    };
      
  /* Test all pattern sets */

  /* Training patterns */
  Error = 0; Anfis->Error = 0;
  for (j = 0 ; j < Anfis->out; j++)
    errors[j] = 0;
  printf ("Now propagating training patterns.\n");
  for (i = 0; i < TrainingPattern->count; i++)
    {
      x = TSK_Propagate_Pattern (Anfis, TrainingPattern, i);
      Error += x;
      for (j = 0 ; j < Anfis->out; j++)
	errors[j] += sqr(Anfis->errors[j]);
    };
  printf ("Overall Root Mean Square Error: %f.\n", 
	  sqrt(Error / (TrainingPattern->count * Anfis->out)));
  for (j = 0; j < Anfis->out; j++)
    {
      printf ("Output %i, Non-Dimensional Error Index: %f.\n", j,
	      sqrt(errors[j]/TrainingPattern->count) / 
	      TrainingPattern->pstddev[Anfis->in + j]);
    };
  
  printf ("Write training result file %s.\n", LP.trainresultname);
  TSK_Test_Pattern (Anfis, TrainingPattern, LP.trainresultname);
      
  
  /* Generalization patterns */
  if (GeneralizationPattern != NULL)
    {
      Error = 0; Anfis->Error = 0;
      for (j = 0 ; j < Anfis->out; j++)
	errors[j] = 0;
      printf ("Now propagating generalization patterns.\n");
      for (i = 0; i < GeneralizationPattern->count; i++)
	{
	  x = TSK_Propagate_Pattern (Anfis, GeneralizationPattern, i);
	  Error += x;
	  for (j = 0 ; j < Anfis->out; j++)
	    errors[j] += sqr(Anfis->errors[j]);
	};
      printf ("Overall Root Mean Square Error: %f.\n", 
	      sqrt(Error / (GeneralizationPattern->count * Anfis->out)));
      for (j = 0; j < Anfis->out; j++)
	{
	  printf ("Output %i, Non-Dimensional Error Index: %f.\n", j,
		  sqrt(errors[j]/GeneralizationPattern->count) / 
		  GeneralizationPattern->pstddev[Anfis->in + j]);
	};
   
      printf ("Write generalization result file %s.\n", LP.genresultname);
      TSK_Test_Pattern (Anfis, GeneralizationPattern, LP.genresultname);
    };

  /* Test patterns */
  if (TestPattern != NULL)
    {
      Error = 0; Anfis->Error = 0;
      for (j = 0 ; j < Anfis->out; j++)
	errors[j] = 0;
      printf ("Now propagating test patterns.\n");
      for (i = 0; i < TestPattern->count; i++)
	{
	  x = TSK_Propagate_Pattern (Anfis, TestPattern, i);
	  Error += x;
	  for (j = 0 ; j < Anfis->out; j++)
	    errors[j] += sqr(Anfis->errors[j]);
	};
      printf ("Overall Root Mean Square Error: %f.\n", 
	      sqrt(Error / (TestPattern->count * Anfis->out)));
      for (j = 0; j < Anfis->out; j++)
	{
	  printf ("Output %i, Non-Dimensional Error Index: %f.\n", j,
		  sqrt(errors[j]/TestPattern->count) / 
		  TestPattern->pstddev[Anfis->in + j]);
	};
      
      printf ("Write test result file %s.\n", LP.testresultname);
      TSK_Test_Pattern (Anfis, TestPattern, LP.testresultname);
    };

  return (epochs);

}; /* TSK Training Cycle */

/**************************************************************************/
/**************************************************************************/
/**************************************************************************/

void Mamdani_Test(const char *filename)
{
  int i,j;
  double Error;
  double *errors;
  double x;

  /* allocate memory for output errors */
  if ((errors = (double *) calloc (TestPattern->out, sizeof (double))) 
      == NULL)
    {
      printf ("Cannot allocate memory for errors.\n");
      exit (-1);
    };

  Error = 0; Nefprox->Error = 0;
  for (j = 0 ; j< TestPattern->out; j++)
    errors[j] = 0;
  printf ("Now propagating test patterns.\n");
  for (i = 0; i < TestPattern->count; i++)
    {
      x = Mamdani_Propagate_Pattern (Nefprox, TestPattern, i);
      Error += x;
      for (j = 0 ; j < TestPattern->out; j++)
	errors[j] += sqr(Nefprox->errors[j]);
	};
  
  printf ("Overall Root Mean Square Error: %f.\n", 
	  sqrt (Error / (TestPattern->count * Nefprox->out)));
  for (j = 0; j < TestPattern->out; j++)
    {
      errors[j] = sqrt(errors[j] / TestPattern->count) / 
	TestPattern->pstddev[TestPattern->in + j];
      
      printf ("Output %i, Non-Dimensional Error Index: %f.\n", 
	      j, errors[j]);
	};
  printf ("Write test result file %s.\n", filename);
  Mamdani_Test_Pattern (Nefprox, TestPattern, filename);

  return;
}; /* Mamdani_Test */

/**************************************************************************/

void TSK_Test(const char *filename)
{
  int i,j;
  double Error;
  double *errors;
  double x;

  /* allocate memory for output errors */
  if ((errors = (double *) calloc (TestPattern->out, sizeof (double))) 
      == NULL)
    {
      printf ("Cannot allocate memory for errors.\n");
      exit (-1);
    };

  Error = 0; Anfis->Error = 0;
  for (i = 0 ; i < TestPattern->out; i++)
    errors[i] = 0;
  printf ("Now propagating test patterns.\n");
  for (i = 0; i < TestPattern->count; i++)
    {
      x = TSK_Propagate_Pattern (Anfis, TestPattern, i);
      Error += x;
      for (j = 0 ; j < TestPattern->out; j++)
	errors[j] += sqr(Anfis->errors[j]);
    };
  printf ("Overall Root Mean Square Error: %f.\n", 
	  sqrt(Error / (TestPattern->count * Anfis->out)));
  for (j = 0; j < TestPattern->out; j++)
    {
      printf ("Output %i, Non-Dimensional Error Index: %f.\n", j,
	      sqrt(errors[j]/TestPattern->count) / 
	      TestPattern->pstddev[Anfis->in + j]);
    };
      
  printf ("Write test result file %s.\n", filename);
  TSK_Test_Pattern (Anfis, TestPattern, filename);
  return;
}; /* TSK_Test */

/**************************************************************************/

void Mamdani_Gnuplot(const char *filename)
{
  int i;
  char mode[3];

  printf ("Write gnuplot command file %s to plot the fuzzy sets.\n",
	  filename);
  strcpy (mode, "w");
  for (i = 0; i < Nefprox->in; i++)
    {
      Write_Partition_For_Gnuplot (filename, Nefprox->invars[i], mode); 
      strcpy (mode, "a");
    };
  
  for (i = 0; i < Nefprox->out; i++)
    {
      Write_Partition_For_Gnuplot (filename, Nefprox->outvars[i], "a");
    };
  return;
}; /* Mamdani_Gnuplot */

/**************************************************************************/

void TSK_Gnuplot(const char *filename)
{
  int i;
  char mode[3];

  printf ("Write gnuplot command file %s to plot the fuzzy sets.\n",
	  filename);
  strcpy (mode, "w");
  for (i = 0; i < Anfis->in; i++)
    {
      Write_Partition_For_Gnuplot (filename, Anfis->invars[i], mode); 
      strcpy (mode, "a");
    };
      
  return;
}; /* TSK_Gnuplot */

/**************************************************************************/

void usage (const char *command)
{
  printf("Usage\n");
  printf("=====\n");
  printf("Learning: %s -l <command-file>\n", command);
  printf("Testing : %s -t <net-file> <data-file> [<result-file>]\n", 
	 command);
  printf("Create gnuplot file from network: %s -g <net-file> [<gnu-file>]\n", 
	 command);
  return;
}; /* usage */

/**************************************************************************/

int main(argc, argv)
int argc;
char **argv;
{
  int mode;
  char filename[80];

  printf ("NFIDENT - Neuro-Fuzzy Function Approximation\n");
  printf (VERSION"\n");
  printf ("Copyright (C) 1997-2000, Detlef Nauck (nauck@ieee.org)\n");
  printf ("This program may be used and distributed in accordance with the\n");
  printf ("GNU General Public Licence. NO WARRANTY.\n\n");
  Init_Learn_Parameters (&LP);
  Nefprox = NULL;
  Anfis = NULL;
  BestNefprox = NULL;
  BestAnfis = NULL;
  TrainingPattern = NULL;
  GeneralizationPattern = NULL;
  TestPattern = NULL;
  StopOnSignal = 0;

  if ((argc < 2) || argv[1][0] != '-')
    {
      usage(argv[0]);
      return -1;
    }
  else
    {
      switch (argv[1][1])
	{
	  case 'l': 
	    if (0 != Read_Files (argv[2], &LP))
	      {
		return -1;
	      }
	    else
	      {
		mode = 1;
	      }
	    break;
	case 't':
	  if (0 != Read_Net_and_Data (argv[2], argv[3], &LP))
	      {
		return -1;
	      }
	  else 
	    {
	      mode = 2;
	      if (argc == 5)
		{
		  strcpy (filename, argv[4]);
		}
	      else
		{
		  strcpy (filename, "result.dat");
		};
	    };
	  break;
	case 'g':
	  if (0!= Read_Net(argv[2], &LP))
	      {
		return -1;
	      }
	  else 
	    {
	      mode = 3;
	      if (argc == 4)
		{
		  strcpy (filename, argv[3]);
		}
	      else
		{
		  strcpy (filename, "gnu.cmd");
		};
	    };
	  break;
	default: 
	  usage(argv[0]);
	  return -1;
	}/* switch */
    };

  /* set signal handler to handle Ctrl-C during training */
  signal (SIGINT, INTHandler);

  switch (mode)
    {
    case 1:
      if (LP.fstype == MAMDANI)
	{
	  Mamdani_Training_Cycle();
	}
      else
	{
	  TSK_Training_Cycle();
	};
      break;
    case 2:
      if (Nefprox != NULL)
	{
	  Mamdani_Test(filename);
	}
      else
	{
	  TSK_Test(filename);
	};
      break;
    case 3:
      if (Nefprox != NULL)
	{
	  Mamdani_Gnuplot(filename);
	}
      else
	{
	  TSK_Gnuplot(filename);
	};
      break;
    }; /* switch mode */
  printf ("Done.\n");
}
