/* NFIDENT: a program for neuro-fuzzy function approximation

   Copyright (C) 1997-2000 

   Detlef Nauck (nauck@ieee.org, 
   http:\\fuzzy.cs.uni-magdeburg.de/~nauck)  
   hereby disclaims all copyright interests in the program 
   NFIDENT written by Detlef Nauck

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 1, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/**************************************************************************/
/* pattern.c, D.Nauck, 08.Sep.96                                          */
/*------------------------------------------------------------------------*/
/* functions to read and process patterns                                 */
/**************************************************************************/

#include <stdio.h>
#include <math.h>
#include <limits.h>
#include "global.h"

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/* ReadPattern: allocate memory and read patterns from file               */
/*                                                                        */
/* FileName: name of pattern file                                         */
/* NofIn: number of input features                                        */
/* NofOut: number of outputs                                              */
/* NofPat: number of patterns in file                                     */
/* Pats: pointer to a matrix of patterns                                  */
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

int Read_Pattern (const char *FileName, PSet *Pattern)
{
  FILE *PatternFile;
  int i,j;
  int twopercent;

  if ((PatternFile = fopen(FileName, "r")) == NULL)
    {
      printf ("Could not open pattern file %s.\n",FileName);
      return 1;
    };

  /* read number of patterns, number of inputs and outputs*/
  fscanf (PatternFile, "%d%d%d", &(Pattern->count), &(Pattern->in), 
	  &(Pattern->out));
  
  printf ("Reading %i patterns from %s.\n", Pattern->count, FileName);
  twopercent = 1 + Pattern->count / 50;

  /* allocate row pointers*/
  if ((Pattern->Pats = 
       (matrix *) calloc (Pattern->count, sizeof(double *)))==NULL)
    {
      printf("Cannot allocate memory for patterns.\n");
      return -1;
    };
  
  /* allocate max and min vectors */
  if ((Pattern->pmin = (double *) calloc (Pattern->in + Pattern->out, 
					 sizeof (double))) 
      == NULL)
    {
      printf ("Cannot allocate memory for lower boundaries.\n");
      return -2;
    };

  for (i = 0; i < Pattern->in + Pattern->out; i++)
    Pattern->pmin[i] = INT_MAX;

  if ((Pattern->pmax = (double *) calloc (Pattern->in + Pattern->out, 
					 sizeof (double))) 
      == NULL)
    {
      printf ("Cannot allocate memory for upper boundaries.\n");
      return -3;
    };

  for (i = 0; i < Pattern->in + Pattern->out; i++)
    Pattern->pmax[i] = INT_MIN;

  if ((Pattern->pmean = (double *) calloc (Pattern->in + Pattern->out, 
					  sizeof (double))) 
      == NULL)
    {
      printf ("Cannot allocate memory for means.\n");
      return -3;
    };

  if ((Pattern->pstddev = (double *) calloc (Pattern->in + Pattern->out, 
					    sizeof (double))) 
      == NULL)
    {
      printf ("Cannot allocate memory for standard deviations.\n");
      return -3;
    };

  j = 0;

  for (i = 0; i < Pattern->in + Pattern->out; i++)
    {
      Pattern->pmean[i] = 0;
      Pattern->pstddev[i] = 0;
    };

  while (!feof(PatternFile) && (j < Pattern->count))
    {
      /* allocate a new row */
      if ((Pattern->Pats[j] = 
	   (double *) calloc(Pattern->in + Pattern->out, sizeof(double))) 
	  == NULL)
	{
	  printf ("Cannot allocate memory for new pattern.\n");
	  return -(j+1);
	};
  
      /* read patterns into row */
      for (i = 0; i < Pattern->in + Pattern->out; i++)
        {
	  fscanf (PatternFile, "%lf", &(Pattern->Pats[j][i])); 
	  if (Pattern->Pats[j][i] < Pattern->pmin[i])
	    Pattern->pmin[i] = Pattern->Pats[j][i];
	  if (Pattern->Pats[j][i] > Pattern->pmax[i])
	    Pattern->pmax[i] = Pattern->Pats[j][i];
	  Pattern->pmean[i] += Pattern->Pats[j][i];
	};

      j++;
      if ((j % twopercent) == 0)
	{
	  printf (".");
	  fflush (stdout);
	};
    };

  Pattern->count = j; /* if there are less patterns than expected */
  printf (" %i patterns found, done.\n", j);
  fclose(PatternFile);
  
  for (i = 0; i < Pattern->in + Pattern->out; i++)
    {
      Pattern->pmean[i] /= Pattern->count;
    };

  for (j = 0; j < Pattern->count; j++)
    {
      for (i = 0; i < Pattern->in + Pattern->out; i++)
	{
	  Pattern->pstddev[i] += 
	    sqr((Pattern->pmean[i] - Pattern->Pats[j][i]));
	};
    };

  printf ("\nPattern statistics:\n\n");
  for (i = 0; i < Pattern->in + Pattern->out; i++)
    {
      Pattern->pstddev[i] = sqrt(Pattern->pstddev[i] / Pattern->count);
      printf ("%2i) Mean = %lf, StdDev = %lf\n", i, Pattern->pmean[i],
	      Pattern->pstddev[i]);
    };

  return 0;
} /* Read_Pattern */

void Normalize (PSet *P)
{

}; /* Normalize */
