/* NFIDENT: a program for neuro-fuzzy function approximation

   Copyright (C) 1997-2000 

   Detlef Nauck (nauck@ieee.org, 
   http:\\fuzzy.cs.uni-magdeburg.de/~nauck)  
   hereby disclaims all copyright interests in the program 
   NFIDENT written by Detlef Nauck

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 1, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <float.h>
#include "global.h"
#include "defuzz.h"


TSKFuzzySystem *Create_TSK_Fuzzy_System (int nin, int nout, LParam *LP)
{
  TSKFuzzySystem *TSK;

  /* allocate memory for fuzzy system */
  if ((TSK = (TSKFuzzySystem *) calloc (1, sizeof(TSKFuzzySystem))) == NULL)
    {
      printf ("Cannot allocate memory for fuzzy system!\n");
      return NULL;
    };

  /* allocate memory for input variables */
  if ((TSK->invars = (PLingVar *) calloc (nin, sizeof (PLingVar))) == NULL)
    {
      printf ("Cannot allocate memory for input variables!\n");
      return NULL;
    };

  /* allocate memory for error values */
  if ((TSK->errors = (double *) calloc (nout, sizeof (double))) == NULL)
    {
      printf ("Cannot allocate memory for error variables!\n");
      return NULL;
    }
  
  /* allocate memory for output variables */
  if ((TSK->outputs = (double *) calloc (nout, sizeof (double))) == NULL)
    {
      printf ("Cannot allocate memory for output vector!\n");
      return NULL;
    }  
  
  /* allocate memory for the rulebase (MAXRULES pointers to rules) */
  if ((TSK->rulebase = (PTSKRule *) calloc (MAXRULES, 
					  sizeof (PTSKRule))) == NULL)
    {
      printf ("Cannot allocate memory for rulebase!\n");
      return NULL;
    };
  
  TSK->in = nin;
  TSK->out = nout;
  TSK->rules = 0;
  TSK->Error = 0;

  switch (LP->tnorm)
    {
    case TMIN: 
      TSK->tnorm = minimum;
      break;
    case TPROD:
      TSK->tnorm = prod;
      break;
    default:
      TSK->tnorm = prod;
    };

  return TSK;
}; /* Create TSK_Fuzzy_System */

PTSKRule Allocate_New_TSK_Rule (int nin, int nout)
{
  PTSKRule rule;
  int i;

  if ((rule = (PTSKRule) calloc (1, sizeof (TSKRule))) == NULL)
    {
      printf ("Cannot allocate memory for a new rule.\n");
      return NULL;
    };
  if ((rule->antecedent = (int *) calloc(nin, sizeof(int))) == NULL)
    {
      printf ("Cannot allocate memory for a new rule.\n");
      return NULL;
    };

  /* One pointer to an array of doubles for each output variable */
  if ((rule->conclusion = (matrix *) calloc(nout, sizeof(double *))) == NULL)
    {
      printf ("Cannot allocate memory for a new rule.\n");
      return NULL;
    };
  
  for (i = 0; i < nout; i++)
    {
      /* One coefficient for each input variable plus a constant */
      if ((rule->conclusion[i] = (double *) calloc(nin+1, 
						  sizeof(double))) == NULL)
	{
	  printf ("Cannot allocate memory for a new rule.\n");
	  return NULL;
	};
    };

  /* Now the same for the delta values of the conclusion parameters: */
  /* One pointer to an array of doubles for each output variable */
  if ((rule->delta = (matrix *) calloc(nout, sizeof(double *))) == NULL)
    {
      printf ("Cannot allocate memory for a new rule.\n");
      return NULL;
    };
  
  for (i = 0; i < nout; i++)
    {
      /* One coefficient for each input variable plus a constant */
      if ((rule->delta[i] = (double *) calloc(nin+1, sizeof(double))) == NULL)
	{
	  printf ("Cannot allocate memory for a new rule.\n");
	  return NULL;
	};
    };

  /* And now the same for the old delta values of the conclusion parameters: */
  /* One pointer to an array of doubles for each output variable */
  if ((rule->old_delta = (matrix *) calloc(nout, sizeof(double *))) == NULL)
    {
      printf ("Cannot allocate memory for a new rule.\n");
      return NULL;
    };
  
  for (i = 0; i < nout; i++)
    {
      /* One coefficient for each input variable plus a constant */
      if ((rule->old_delta[i] = (double *) calloc(nin+1, sizeof(double))) 
	  == NULL)
	{
	  printf ("Cannot allocate memory for a new rule.\n");
	  return NULL;
	};
    };

  /* allocate memory for the rule outputs */
  if ((rule->output = (double *) calloc(nout, sizeof(double))) == NULL)
    {
      printf ("Cannot allocate memory for a new rule.\n");
      return NULL;
    };
  
  return rule;
};


void Compute_TSK_Rule_Fulfillments (TSKFuzzySystem *TSK)
{
  int i;
  int j;
  double hit;
  PTSKRule rule;
  int *antec;
  int *concl;
  double hitsum;

  hitsum = 0;
  for (i = 0; i < TSK-> rules; i++)
    {
      hit = 1;
      rule = TSK->rulebase[i];
      antec = rule->antecedent;

      for (j = 0; j < TSK->in; j++)
	{
	  if (antec[j] >= 0)
	    hit = TSK->tnorm(hit, TSK->invars[j]->fsets[antec[j]].hit);
	};
      rule->hit = hit;
      hitsum += hit;
      rule->normhit = hit;
    };

  if (hitsum > 0)
    {
      for (i = 0; i < TSK-> rules; i++)
	{
	  TSK->rulebase[i]->normhit /= hitsum;
	};
    };
}; /* Compute TSK Rule Fulfillments */

/* compute the final output values by defuzzification */
void Compute_TSK_Output (TSKFuzzySystem *TSK, PSet *P, int which)
{
  int i, j, k;
  double summand;
  double nominator;
  double denominator;
  PTSKRule rule;

  for (i = 0; i < TSK->out; i++)
    {
      TSK->outputs[i] = 0; 
      for (j = 0; j < TSK->rules; j++)
	{
	  rule = TSK->rulebase[j];
	  if (rule->hit > 0)
	    {
	      summand = 0; 
	      for (k = 0; k < TSK->in; k++)
		{
		  summand += P->Pats[which][k] * rule->conclusion[i][k];
		};
	      rule->output[i] = (summand + rule->conclusion[i][TSK->in]);
	      TSK->outputs[i] += rule->output[i] * rule->normhit;
	    } /* if */
	  else
	    {
	      rule->output[i] = 0;
	    };
	}; /* for j */
    }; /* for i */
}; /* Compute TSK Output */

double Compute_TSK_Error (TSKFuzzySystem *TSK, PSet *P, int which)
{
  int i;
  double error;

  error = 0;
  for (i = 0; i < P->out; i++)
    {
      TSK->errors[i] = P->Pats[which][P->in + i] - TSK->outputs[i];
      error += sqr(TSK->errors[i]);
    };
  return error;
  
};/* Compute TSK Error */

double TSK_Propagate_Pattern (TSKFuzzySystem *TSK, PSet *P, int which)
{
  double error;

  Compute_Pattern_Memberships (TSK->invars, P->Pats[which], 0, P->in);

  Compute_TSK_Rule_Fulfillments (TSK);

  Compute_TSK_Output (TSK,P,which);

  error = Compute_TSK_Error (TSK, P, which);
  TSK->Error += error;
  return error;

}; /* TSK Propagate Pattern */


int Create_TSK_Rulebase (TSKFuzzySystem *TSK, PSet *P, LParam *LP)
{
  int i,j,k,l;
  int imaxfs[MAXIN];
  int index;
  int equal;
  double max;
  double hit;
  double output;
  double *pats;
  matrix *mean;
  int start;

  start = TSK->rules;

  if ((mean = (matrix *) calloc(MAXRULES, sizeof(double *))) == NULL)
    {
      printf ("Cannot allocate memory for mean outputs.\n");
      exit (-1);
    };
  for (i = 0; i < MAXRULES; i++)
    {
      if ((mean[i] = (double *) calloc(TSK->out, sizeof(double))) == NULL)
	{
	  printf ("Cannot allocate memory for mean outputs.\n");
	  exit (-1);
	};
    };

  if ((pats = (double *) calloc(MAXRULES, sizeof(double))) == NULL)
    {
      printf ("Cannot allocate memory for mean outputs.\n");
      exit (-1);
    };


  for (i = 0; i < P->count; i++)
    { /* process all patterns */
      Compute_Pattern_Memberships (TSK->invars, P->Pats[i], 0, P->in);

      for (j = 0; j < TSK->in; j++)
	{ /* find best antecedent for the current pattern */
	  max = 0;
	  index = -1;
	  for (k = 0; k < TSK->invars[j]->count; k++)
	    {
	      if (max < (hit = TSK->invars[j]->fsets[k].hit))
		{
		  max = hit;
		  index = k;
		};
	    }; /* for k */
	  if (index < 0)
	    {
	      printf ("Error in fuzzy partition of input %i.\n", j);
	      return 0;
	    };
	  imaxfs[j] = index;
	}; /* for j */

      equal = 0; j = 0;
      while (!equal && (j < TSK->rules))
	{ /* check, whether there is already such an antecedent */
	  k = 0; equal = 1;
	  while (equal && (k < TSK->in))
	    {
	      equal &= (imaxfs[k] == TSK->rulebase[j]->antecedent[k]);
	      k++;
	    };
	  if (equal)
	    {
	      for (l = 0; l < TSK->out; l++)
		{
		  output = 0;
		  for (k = 0; k < TSK->in; k++)
		    {
		      output += P->Pats[i][k] * 
			TSK->rulebase[j]->conclusion[l][k];
		    };
		  mean[j][l] += P->Pats[i][P->in+l] - output;
		};
	      pats[j]++;
	    };
	  j++;
	}; /* while (!equal) */

      if ((!equal) && (TSK->rules < MAXRULES))
	{ /* there is no such antecedent yet, so lets create a new rule, */
	  /* if we are still allowed to do this;      */

	  /* allocate a new rule, and set antecedent */
	  TSK->rulebase[TSK->rules] = Allocate_New_TSK_Rule(TSK->in, TSK->out);
	  for (j = 0; j < TSK->in; j++)
	    {
	      TSK->rulebase[TSK->rules]->antecedent[j] = imaxfs[j];
	    };
	  /* now compute suitable coeffiecients for each output variable, */
	  /* if LMSE is not used, else the coefficients are zero          */
	  if (LP->uselmse == 0)
	    {
	      for (j = 0; j < TSK->out; j++)
		{
		  output = 0;
		  for (k = 0; k < TSK->in; k++)
		    { /* select coefficients randomly, and compute output */
		      output += P->Pats[i][k] * 
			(TSK->rulebase[TSK->rules]->conclusion[j][k] = 
			 LP->tskinit * (2.0 * ((double) rand()) / 
					RAND_MAX - 1));
		    };
		  mean[TSK->rules][j] += P->Pats[i][P->in+j] - output;
		}; /* for j */
	      pats[TSK->rules]++;
	    }; /* if */

	  TSK->rules++;
	}; /* if (!equal) */

    }; /* for i (end of pattern loop) */

  /* now compute the constant such that the mean output is obtained, */
  /* but only for the new rules, and if LMSE is not used */
  if (LP->uselmse == 0)
    {
      for (i = start; i < TSK->rules; i++)
	{
	  for (j = 0; j < TSK->out; j++)
	    {
	      TSK->rulebase[i]->conclusion[j][TSK->in] = 
		mean[i][j] / pats[i];
	    };
	};
    };

  for (i = 0; i < MAXRULES; i++)
    {
      free (mean[i]);
    };
  free (mean);
  free (pats);

  return TSK->rules;
}; /* Create TSK Rulebase*/

int TSK_Test_Pattern (TSKFuzzySystem *TSK, PSet *P, const char *filename)
{
  int i,j;
  FILE *resultfile;


  if ((resultfile = fopen (filename, "w")) == NULL)
    {
      printf ("Cannot open result file.\n");
      return -1;
    };
  
  for (i = 0; i < P->count; i++)
    {
      TSK_Propagate_Pattern (TSK, P, i);
      for (j = 0; j < P->out; j ++)
	{
	  fprintf (resultfile, "%10.4f %10.4f %10.4f", P->Pats[i][P->in+j],
		   TSK->outputs[j],  P->Pats[i][P->in+j] - TSK->outputs[j]);
	};
      fprintf (resultfile, "\n");
    };
  fclose (resultfile);
  return 0;
}; /* TSK_Test_Pattern */


void Copy_TSK (TSKFuzzySystem *TSK, TSKFuzzySystem **Copy, LParam *LP)
{
  int i, j, k;
  TSKFuzzySystem *TC;

  if (TSK != NULL)
    {
      TC = *Copy;
      if (TC == NULL)
	{
	  TC = Create_TSK_Fuzzy_System (TSK->in, TSK->out, LP);
	  TC->tnorm = TSK->tnorm;

	  for (i = 0; i < TSK->in; i++)
	    {
	      if ((TC->invars[i] = 
		   (LingVar *) calloc (1, sizeof(LingVar))) == NULL)
		{
		  printf ("Failed to copy %ith input variable.\n", i);
		  exit (-(i+1));
		};
	      CopyLV (TSK->invars[i], TC->invars[i]);
	    };

	  for (i = 0; i < TSK->rules; i++)
	    {
	      TC->rulebase[i] = Allocate_New_TSK_Rule(TC->in, TC->out);
	      for (j = 0; j < TSK->in; j++)
		{
		  TC->rulebase[i]->antecedent[j] = 
		    TSK->rulebase[i]->antecedent[j];
		};
	      for (j = 0; j < TSK->out; j++)
		{
		  for (k = 0; k < TSK->in + 1; k++)
		    { 
		      TC->rulebase[i]->conclusion[j][k] = 
		      TSK->rulebase[i]->conclusion[j][k];
		    };
		}; /* for j */
	      TC->rules++;
	    }; /* for i */

	  *Copy = TC;

	}/* creation of new TSKFuzzySystem completed */

      else 
	{ /* there is already a copy, just update parameters */
	  for (i = 0; i < TSK->in; i++)
	    {
	      for (j = 0; j < TSK->invars[i]->count; j++)
		{
		  TC->invars[i]->fsets[j].a = TSK->invars[i]->fsets[j].a;
		  TC->invars[i]->fsets[j].b = TSK->invars[i]->fsets[j].b;      
		  TC->invars[i]->fsets[j].c = TSK->invars[i]->fsets[j].c;
		  TC->invars[i]->fsets[j].d = TSK->invars[i]->fsets[j].d;
		};
	    };
	  for (i = 0; i < TSK->rules; i++)
	    {
	      for (j = 0; j < TSK->out; j++)
		{
		  for (k = 0; k < TSK->in + 1; k++)
		    { 
		      TC->rulebase[i]->conclusion[j][k] = 
		      TSK->rulebase[i]->conclusion[j][k];
		    };
		};
	    };
	}; /* update of existing copy completed */
    };
} /* Copy_TSK*/

